#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from pathlib import Path
from collections import UserDict
import zipfile, shutil, xml.sax, os, os.path, re, mimetypes, time

import appy.pod
from appy import utils
from appy.xml import Tag
from appy.pod import PodError
from appy.utils.zip import unzip, zip
from appy.pod.buffers import FileBuffer
from appy.utils.path import FolderDeleter
from appy.pod.converter import FILE_TYPES
from appy.pod import styles_manager as sm
from appy.pod import doc_importers as importers
from appy.pod.xhtml2odt import Xhtml2OdtConverter
from appy.pod.pod_parser import PodParser, PodEnvironment, OdInsert

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
BAD_CTX     = 'Context must be either a dict, a UserDict or an instance.'
RES_FOLDER  = 'Result file "%s" is a folder, please remove this.'
RES_EXISTS  = 'Result file "%s" exists.'
TEMP_W_KO   = 'I cannot create temp folder "%s". %s'
NO_PY_PATH  = 'Extension of result file is "%s". In order to perform ' \
              'conversion from ODT to this format we need to call ' \
              'LibreOffice. But the Python interpreter which runs the ' \
              'current script does not know UNO, the library that allows to ' \
              'connect to LibreOffice in server mode. If you can\'t install ' \
              'UNO in this Python interpreter, you can specify, in parameter ' \
              '"pythonWithUnoPath", the path to a UNO-enabled Python ' \
              'interpreter. One such interpreter may be found in ' \
              '<libreoffice_path>/program.'
PY_PATH_KO  = '"%s" is not a file. You must here specify the absolute ' \
              'path of a Python interpreter (.../python, .../python.sh, ' \
              '.../python.exe, .../python.bat...).'
R_TYPE_KO   = 'Result "%s" has a wrong extension. Allowed extensions are: "%s".'
CONV_ERROR  = 'An error occurred during the conversion. %s'
BAD_LO_PORT = 'Wrong LibreOffice port "%s". Make sure it is an integer.'
XHTML_ERROR = 'An error occurred while rendering XHTML content.'
WARN_OD_KO  = 'Warning: your OpenDocument file may not be complete (ie ' \
              'imported documents may not be present). This is because we ' \
              'could not connect to LibreOffice in server mode: %s'
DOC_KO      = 'Please specify a document to import, either with a stream ' \
              '(parameter "content") or with a path (parameter "at").'
DOC_FMT_KO  = 'POD was unable to deduce the document format. Please specify ' \
              'it through parameter named "format" (=odt, gif, png, ...).'
DOC_FMT_NS  = 'Format "%s" is not supported.'
WARN_FIN_KO = 'Warning: error while calling finalize function. %s'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Make sure any file on disk is opened as UTF-8. Windows' default encoding is
# still CP1252 :-D
enc = {'encoding': 'utf-8'}

# Default styles added by POD in content.xml and styles.xml
POD_STYLES = {}
podFolder = os.path.dirname(appy.pod.__file__)
for name in ('content', 'styles'):
    f = open('%s/%s.xmlt' % (podFolder, name), **enc)
    POD_STYLES[name] = f.read()
    f.close()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class CsvOptions:
    '''When the POD result is of type CSV, use this class to define CSV
       options.'''

    # Available encodings for CSV output files
    encodings = {'utf-8': 76, 'latin-1': 12}

    def __init__(self, fieldSeparator=';', textDelimiter='"', encoding='utf-8'):
        # The delimiter used between every field on a single line
        self.fieldSeparator = fieldSeparator
        # The character used to "wrap" every value. Can be the empty string.
        self.textDelimiter = textDelimiter
        # The ODT result's file encoding. Defaults is "utf-8". "latin-1" is also
        # allowed, or any numeric value as defined at the URL mentioned in
        # appy/pod/converter.py, at variable HELP_CSV_URL.
        self.encoding = encoding

    def asString(self):
        '''Returns options as a string, ready-to-be sent to the Converter'''
        # Manage the optional text delimiter
        delim = ord(self.textDelimiter) if self.textDelimiter else ''
        # Manage encoding
        enc = self.encoding
        if isinstance(enc, int) or enc.isdigit():
            encoding = enc
        else:
            encoding = CsvOptions.encodings.get(enc) or 76
        return '%s,%s,%s,1' % (ord(self.fieldSeparator), delim, encoding)

# Create a default instance
CsvOptions.default = CsvOptions()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Renderer:
    '''Produce (="render") a document from a template'''

    templateTypes = ('odt', 'ods') # Types of POD templates
    # Regular expression for finding the title within metadata
    metaRex = re.compile('<dc:title>.*?</dc:title>', re.S)
    metaHook = '</office:meta>'

    # Make some elements directly available here
    CsvOptions = CsvOptions

    # ODF tags that may hold POD expressions
    allExpressionsHolders = (
      'if',      # Conditional fields
      'change',  # Track-changed text
      'input',   # Text input fields
      'db'       # Database fields
    )
    defaultExpressionsHolders = ('if', 'change', 'input')

    def __init__(self, template, context, result, pythonWithUnoPath=None,
      ooServer='localhost', ooPort=2002, stream='auto', stylesMapping={},
      html=False, forceOoCall=False, finalizeFunction=None,
      overwriteExisting=False, raiseOnError=False, findImage=None,
      rotateImages=False, stylesTemplate=None, optimalColumnWidths=False,
      distributeColumns=False, script=None, managePageStyles=None,
      resolveFields=False, expressionsHolders=defaultExpressionsHolders,
      metadata=True, pdfOptions='ExportNotes=True', csvOptions=None,
      deleteTempFolder=True, protection=False, pageStart=1, tabbedCR=False,
      fonts=None):
        '''Base on a document template (whose path is in p_template), which is
           an ODT or ODS file containing special expressions and statements
           written in Python, this renderer generates an ODT file (whose path is
           in p_result), that is a copy of the template whose expressions and
           statements have been replaced with the objects defined in the
           p_context.'''

        # If p_result does not end with .odt or .ods, the Renderer will call
        # LibreOffice (LO) to perform a conversion. If p_forceOoCall is True,
        # even if p_result ends with .odt, LO will be called, not for performing
        # a conversion, but for updating some elements like indexes (table of
        # contents, etc) and sections containing links to external files (which
        # is the case, for example, if you use the default function "document").

        # If the Python interpreter which runs the current script is not
        # UNO-enabled, this script will run, in another process, a UNO-enabled
        # Python interpreter (whose path is p_pythonWithUnoPath) which will call
        # LO.

        # When pod needs to communicate with LO, it will assume LO runs on
        # server named p_ooServer, listening on port p_ooPort. If you run
        # several LibreOffice servers, p_ooPort may specify a list of ports. In
        # that case, pod will load-balance requests to the appropriate server.

        # p_stream dictates the communication between the renderer and LO.
        # ----------------------------------------------------------------------
        # p_stream can hold the following values.
        # ----------------------------------------------------------------------
        # "auto"     | if p_ooServer is "localhost", exchanging the ODT result
        # (default)  | produced by the renderer (=the input) and the converted
        #            | file produced by LO (=the output) will be made via files
        #            | on disk. Else, the renderer will consider that LO runs on
        #            | another machine and the exchange will be made via streams
        #            | over the network.
        # ----------------------------------------------------------------------
        # True       | If you want to force communication of the input and
        #            | output files as streams, use stream=True.
        # ----------------------------------------------------------------------
        # False      | If you want to force communication of the input and
        #            | output files via the disk, use stream=False.
        # ----------------------------------------------------------------------
        # "in"       | The input is streamed and the output is written on disk.
        # ----------------------------------------------------------------------
        # "out"      | The input is read from disk and the output is streamed.
        # ----------------------------------------------------------------------

        # If you plan to make "XHTML to OpenDocument" conversions (via the POD
        # function "xhtml"), you may specify a "styles mapping" in
        # p_stylesMapping. Moreover, if the input(s) to this function is valid
        # HTML but not valid XHTML, specify p_html being True.

        # If you specify a function in p_finalizeFunction (or a list/tuple of
        # functions), it will be called by the renderer before re-zipping the
        # ODT/S result. This way, you can still perform some actions on the
        # content of the ODT/S file before it is zipped and potentially
        # converted. Every such function must accept 2 args:
        #  *  the absolute path to the temporary folder, containing the
        #     un-zipped content of the ODT/S result;
        #  *  the Renderer instance.

        # If you set p_overwriteExisting to True, the renderer will overwrite
        # the result file. Else, an exception will be thrown if the result file
        # already exists.

        # If p_raiseOnError is False (the default value), any error encountered
        # during the generation of the result file will be dumped into it, as a
        # Python traceback within a note. Else, the error will be raised.

        # When, in the process of converting a chunk of XHTML code to ODF, a
        # "img" tag is encountered, POD performs a HTTP GET to retrieve it. If
        # the image is stored on the same server running POD, this can be
        # problematic in two ways:
        # 1) regarding performance, it would be better to retrieve the image
        #    from disk;
        # 2) via HTTP, POD may not be allowed to retrieve the image.
        # In order to overcome these problems, place a function in "findImage".
        # While converting chunks of XHTML code, everytime an "img" tag is
        # encountered, this function will be called, with the image URL as
        # unique arg. If the function returns the absolute path to the image, it
        # will be used instead of performing a HTTP GET request. If the function
        # returns None (ie: it has detected that is is a "truly external" URL),
        # the HTTP GET will nevertheless be performed.

        # If p_rotateImages is True, for every image to inject in the result,
        # Appy will call ImageMagick to read EXIF data. If an orientation is
        # defined, ImageMagick will be called to effectively apply the
        # corresponding rotation to the image. Because reading EXIF data on
        # every image to import may take some processing, p_rotateImages is
        # False by default.

        # p_stylesTemplate can be the path to a LibreOffice file (ie, a .ott
        # file) whose styles will be imported within the result.

        # p_optimalColumnWidths corresponds to the homonym option to
        # converter.py, excepted that values "True" or "False" must be boolean
        # values. Note that the POD function "xhtml" requires this parameter to
        # be "OCW_.*" to be fully operational. When p_optimalColumnWidths is not
        # False, forceOoCall is forced to True.

        # p_distributeColumns corresponds to the homonym option to converter.py,
        # excepted that values "True" or "False" must be boolean values. Note
        # that the POD function "xhtml" requires this parameter to be "DC_.*" to
        # be fully operational. When p_distributeColumns is not False,
        # forceOoCall is forced to True.

        # p_script is the absolute path to a Python script containing functions
        # that the converter will call in order to customize the process of
        # manipulating the document via the LibreOffice UNO interface. For more
        # information, see appy/pod/converter.py, option "-s". Note that when
        # such p_script is specified, p_forceOoCall is forced to True.

        # If this document is a sub-document to be included in a master one, it
        # has sense to set a specific value for parameter p_managePageStyles:
        # ----------------------------------------------------------------------
        #  "rename" | will rename all page styles with unique names. This way,
        #           | when imported into a master document, no name clash will
        #           | occur and elements tied to page styles (like headers and
        #           | footers) will be correctly imported for all included
        #           | sub-documents. The "do... pod" statement automatically
        #           | sets this parameter to "rename";
        # ----------------------------------------------------------------------
        #   <int>   | will rename page styles similarly to "rename" and will
        #           | also restart, in the sub-document, page numbering at this
        #           | integer value. It will only work if page styles are
        #           | explicitly applied to pages in the sub-pod template. In
        #           | order to achieve this with LibreOffice, open your
        #           | template, set the cursor somewhere in the first page,
        #           | open the pane with page styles, and double-clic on the
        #           | "default style". To check if it has worked, click inside
        #           | the first paragraph in the page, open its properties and
        #           | check that a page break has been defined on it.
        # ----------------------------------------------------------------------

        # If you want to replace (some) fields with their "hard-coded" values,
        # use p_resolveFields. It can be useful, for instance, if the POD result
        # must be included in a master document (via statement "do ... from
        # pod"), but the total number of pages must be kept as is. If
        # p_resolveFields is True, all fields will be resolved, excepted special
        # ones like PageNumber. Indeed, its value is different from one page to
        # another, so resolving it to a single, hard-coded value has no sense.
        # You can also specify "PageCount" as value for p_resolveField. In this
        # case, only the field(s) containing the total number of pages will be
        # resolved. Use it if it is the only field you need to resolve, it will
        # be more performant. Note that when p_resolveFields is set,
        # p_forceOoCall is forced to True.

        # POD expressions may be defined at various places within a document
        # template. Possible places are listed in static variable
        # "allExpressionsHolders" hereabove. Default places are listed in static
        # variable "defaultExpressionsHolders" hereabove.

        # If p_metadata is True, the result will get a metadata property "title"
        # (in meta.xml) being the result file name, without its extension. If
        # p_metadata is a string, it will be used as title.

        # If the result must be produced in PDF, options being specific to this
        # format can be passed in p_pdfOptions, as a comma-separated string of
        # key=value pairs, as in

        #            ExportNotes=False,PageRange=1-20,Watermark=Sample

        # More info in appy/pod/converter.py.

        # If the result must be produced in CSV, options being specific to this
        # format can be passed in p_csvOptions, as an instance of class
        # CsvOptions, available on class Renderer as Renderer.CsvOptions. If
        # p_csvOptions is None, defauls CSV options will be used from
        # Renderer.CsvOptions.default.

        # If this renderer is called by another one (its "parent"), it may
        # receive an additional parameter, p_deleteTempFolder. If False, the
        # temp folder may be reused by other children of the parent renderer, so
        # we do not delete it (the parent will do it at the end of the whole
        # process).

        # if "protection" is True, you can define variables "cellProtected"
        # and/or "sectionProtected" to determine, respectively, if a cell or a
        # section must be write-protected. It is not enabled by default because
        # it has a little impact on performance and on the size of the p_result
        # (especially large ods results).

        # If p_pageStart is different from 1, the produced document will start
        # page numbering at this number. In that case, p_forceOoCall will be
        # forced to True.

        # By default, the *c*arriage *r*eturn (CR) char is converted to ODF tag
        # <text:line-break/>. This may cause problems with justified text. These
        # problems are overcomed by prefixing the tag with <text:tab/> for
        # producing <text:tab/><text:line-break/>. But beware: "tabbing" the CR
        # that way may cause problems when exporting the POD result to Microsoft
        # Word. This is why attribute "tabbedCR" exists.

        # Attribute "fonts" may hold the name of a font that must exist on the
        # machine running this renderer. If such a name is passed, all base
        # styles defined in the POD template will be redefined with this font.
        # On (most? all?) Linux machines, list the available fonts via a command
        # like:
        #                      fc-list :lang=en family

        self.template = template
        self.templateType = self.getTemplateType(template)
        self.result = result
        self.resultType = os.path.splitext(result)[1].strip('.')
        self.contentXml = None # Content (string) of content.xml
        self.stylesXml = None # Content (string) of styles.xml
        self.manifestXml = None # Content (string) of manifest.xml
        self.metaXml = None # Content (string) of meta.xml
        # Manages the styles defined into the ODT template
        self.tempFolder = None
        self.env = None
        self.pythonWithUnoPath = pythonWithUnoPath
        self.ooServer = ooServer
        self.ooPort = self.manageLoPort(ooPort)
        self.stream = stream
        # p_forceOoCall may be forced to True
        self.forceOoCall = forceOoCall or bool(optimalColumnWidths) or \
          bool(distributeColumns) or bool(script) or bool(resolveFields) or \
          (pageStart > 1)
        # Must the POD post-processor (ppp) be enabled ?
        self.ppp = False
        self.finalizeFunction = self.formatFinalizeFunction(finalizeFunction)
        self.overwriteExisting = overwriteExisting
        self.raiseOnError = raiseOnError
        self.findImage = findImage
        self.rotateImages = rotateImages
        self.stylesTemplate = stylesTemplate
        self.optimalColumnWidths = optimalColumnWidths
        self.distributeColumns = distributeColumns
        self.script = script
        self.managePageStyles = managePageStyles
        self.resolveFields = resolveFields
        self.expressionsHolders = expressionsHolders
        self.metadata = metadata
        self.pdfOptions = pdfOptions
        self.csvOptions = csvOptions
        self.deleteTempFolder = deleteTempFolder
        self.protection = protection
        self.pageStart = pageStart
        self.tabbedCR = tabbedCR
        self.fonts = fonts
        # If sub-renderers are called, keep a trace of them
        self.children = {} # ~{s_templatePath: Renderer}~
        # Keep trace of the original context given to the renderer
        self.originalContext = context
        # Remember potential files or images that will be included through
        # "do ... from document" statements: we will need to declare them in
        # META-INF/manifest.xml. Keys are file names as they appear within the
        # ODT file (to dump in manifest.xml); values are original paths of
        # included images (used for avoiding to create multiple copies of a file
        # which is imported several times).
        self.fileNames = {}
        self.prepareFolders()
        # Unzip the p_template
        info = unzip(template, self.unzipFolder, odf=True)
        self.contentXml = info['content.xml']
        self.stylesXml = info['styles.xml']
        # Manage the styles defined into the ODT template
        self.stylesManager = sm.StylesManager(self)
        # From LibreOffice 3.5, it is not possible anymore to dump errors into
        # the resulting ods as annotations. Indeed, annotations can't reside
        # anymore within paragraphs. ODS files generated with pod and containing
        # error messages in annotations cause LibreOffice 3.5 and 4.0 to crash.
        # LibreOffice >= 4.1 simply does not show the annotation.
        if info['mimetype'].decode() == utils.mimeTypes['ods']:
            self.raiseOnError = True
        # Create the parsers for content.xml and styles.xml
        self.createParsers(context)
        # Store the styles mapping
        self.setStylesMapping(stylesMapping)
        # Store the p_html parameter
        self.html = html

    def reinit(self, result, context):
        '''Re-initialise this renderer (p_self) for recycling him and produce
           another p_result with another p_context.'''
        self.result = result
        self.originalContext = context
        # Re-create POD parsers
        self.createParsers(context)
        # Reinitialise attributes being specific to a given result
        smap = self.stylesManager.stylesMapping
        self.stylesManager = sm.StylesManager(self)
        self.stylesManager.stylesMapping = smap

    # Attributes to clone to a sub-renderer when using m_clone hereafter
    cloneAttributes = ('pythonWithUnoPath', 'ooServer', 'ooPort', 'html',
      'raiseOnError', 'findImage', 'rotateImages', 'stylesTemplate',
      'optimalColumnWidths', 'distributeColumns', 'expressionsHolders',
      'protection', 'tabbedCR', 'fonts')

    def clone(self, template, context, result, **params):
        '''Creates another Renderer instance, similar to p_self, but for
           rendering a different p_result based on another p_context and
           p_template, with possibly different other p_params. Any parameter not
           being in p_params but listed in Renderer.cloneAttributes will be
           copied from p_self to the clone.'''
        # Complete p_params with attributes listed in Renderer.cloneAttributes
        for name in Renderer.cloneAttributes:
            if name not in params:
                params[name] = getattr(self, name)
        # Create and return the clone
        return Renderer(template, context, result, **params)

    def manageLoPort(self, loPort):
        '''Ensure the specified LO port (p_loPort) is valid. If it is a list of
           ports, create the data structure allowing to manage load-balancing
           between LO servers.'''
        port = loPort or 2002
        if not isinstance(port, int):
            raise PodError(BAD_LO_PORT % str(port))
        return port

    def formatFinalizeFunction(self, fun):
        '''Standardize "finalize function" p_fun as a list of functions'''
        if not fun: return
        elif isinstance(fun, list): return fun
        elif isinstance(fun, tuple): return list(fun)
        else: return [fun]

    def enablePpp(self):
        '''Activate the POD post-processor, which is implemented as a series of
           UNO commands.'''
        self.forceOoCall = True
        self.ppp = True

    def getCompleteContext(self, context):
        '''Create and return a complete context, from the one given by the pod
           developer (p_context).'''
        r = {}
        if hasattr(context, '__dict__'):
            r.update(context.__dict__)
        elif isinstance(context, dict) or isinstance(context, UserDict):
            r.update(context)
        else:
            raise PodError(BAD_CTX)
        # Incorporate the default, unalterable, context
        r.update({'xhtml': self.renderXhtml, 'text': self.renderText,
          'document': self.importDocument, 'image': self.importImage,
          'pod': self.importPod, 'cell': self.importCell,
          'shape': self.drawShape, 'TableProperties': sm.TableProperties,
          'BulletedProperties': sm.BulletedProperties,
          'NumberedProperties': sm.NumberedProperties,
          'pageBreak': self.insertPageBreak,
          'columnBreak': self.insertColumnBreak,
          # Variables to use for representing pod-reserved chars
          'PIPE': '|', 'SEMICOLON': ';'})
        # Developer, forget the following line
        if '_ctx_' not in r: r['_ctx_'] = r
        return r

    def createPodParser(self, odtFile, context, inserts=None):
        '''Creates the parser with its environment for parsing the given
           p_odtFile (content.xml or styles.xml). p_context is given by the pod
           user, while p_inserts depends on the ODT file we must parse.'''
        context = self.getCompleteContext(context)
        env = PodEnvironment(context, inserts, self.expressionsHolders)
        fileBuffer = FileBuffer(env, os.path.join(self.tempFolder, odtFile))
        env.currentBuffer = fileBuffer
        return PodParser(env, self)

    def createParsers(self, context):
        '''Creates parsers for content.xml and styles.xml and store them,
           respectively, in p_self.contentParser and p_self.stylesParser.'''
        nso = PodEnvironment.NS_OFFICE
        for name in ('content', 'styles'):
            styleTag = 'automatic-styles' if name == 'content' else 'styles'
            inserts = (OdInsert(POD_STYLES[name], Tag(styleTag, nsUri=nso)),)
            parser = self.createPodParser('%s.xml' % name, context, inserts)
            setattr(self, '%sParser' % name, parser)

    def renderXhtml(self, s, stylesMapping={}, keepWithNext=0,
                    keepImagesRatio=False, imagesMaxWidth='page', html=None,
                    inject=False, unwrap=False):
        '''Method that can be used (under the name "xhtml") into a POD template
           for converting a chunk of XHTML content (p_s) into a chunk of ODF
           content.'''

        # For this conversion, beyond the global styles mapping defined at the
        # renderer level, a specific p_stylesMapping can be passed: any key in
        # it overrides its homonym in the global mapping.

        # Parameter p_keepWithNext is used to prevent the last part of a
        # document to be left alone on top of the last page. Imagine your POD
        # template is an official document ending with some director's scanned
        # signature. Just before this signature, the document body is inserted
        # using m_renderXhtml. Depending on p_s's length, in some cases, the
        # scanned signature may end up alone on the last page. When using
        # p_keepWithNext, POD applies a specific style to p_s's last paragraph,
        # such that it will get standard LibreOffice property "keep-with-next"
        # and will thus be "pushed" on the last page, together with the scanned
        # signature, even if there is still space available on the previous one.

        # p_keepWithNext may hold the following values.
        # ----------------------------------------------------------------------
        #  0  | (the default) Keep-with-next functionality is disabled.
        # ----------------------------------------------------------------------
        #  1  | Keep-with-next is enabled as described hereabove: p_s's last
        #     | paragraph receives LibreOffice property "keep-with-next". It
        #     | also works if the last paragraph is a bulleted or numbered item.
        # ----------------------------------------------------------------------
        #  >1 | If p_keepWithNext is higher than 1, it represents a number of
        #     | characters to "keep-with-next". Indeed, in some cases, keeping
        #     | only the last paragraph may not be sufficient: a bit more text
        #     | could produce a better result. Based on this number of
        #     | characters, POD will determine how many paragraphs will get
        #     | property "keep-with-next" and will apply it to all of them. For
        #     | example, suppose p_keepWithNext is defined to 60. The last 3
        #     | paragraphs contain, respectively, 20, 30 and 35 characters. POD
        #     | will apply property "keep-with-next" to the 2 last paragraphs.
        #     | The algorithm is the following: POD walks p_s's paragraphs
        #     | backwards, starting from the last one, counting and adding the
        #     | number of characters for every walked paragraph. POD continues
        #     | the walk until it reaches (or exceeds) the number of characters
        #     | to keep. When it is the case, it stops. All the paragraphs
        #     | walked so far receive property "keep-with-next".
        #     | ~~~
        #     | POD goes even one step further by applying "keep-with-next"
        #     | properties to tables as well. In the previous example, if, when
        #     | counting 60 characters backwards, we end up in the middle of a
        #     | table, POD will apply property "keep-with-next" to the whole
        #     | table. However, with tables spanning more than one page, there
        #     | is a problem: if property "keep-with-next" is applied to such a
        #     | table, LibreOffice will insert a page break at the beginning of
        #     | the table. This can be annoying. While this may be considered a
        #     | bug (maybe because it represents a constraint being particularly
        #     | hard to implement), it is the behaviour implemented in
        #     | LibreOffice, at least >= 3 and <= 6.4. Consequently, at the POD
        #     | level, an (expensive) workaround has been found: when
        #     | p_keepWithNext characters lead us inside a table, POD will split
        #     | it into 2 tables: a first table containing all the rows that
        #     | were not walked by the keep-with-next algorithm, and a second
        #     | containing the remaining, walked rows. On this second table
        #     | only, property "keep-with-next" is applied. Because splitting
        #     | tables that way requires LibreOffice running in server mode, as
        #     | soon as you specify p_keepWithNext > 1, POD wil assume
        #     | LibreOffice runs in server mode.
        # ----------------------------------------------------------------------

        # If p_keepImagesRatio is True, while importing images from "img" tags
        # within p_s, their width/height ratio will be kept. Note that in most
        # cases, it is useless to specify it, because POD computes the image's
        # real width and height.

        # Parameter p_imagesMaxWidth corresponds to parameter p_maxWith from POD
        # function m_document. Being "page" by default, it prevents the images
        # defined in p_s to exceed the ODT result's page width.
        
        # If p_html is not None, it will override renderer's homonym parameter.
        # Set p_html to True if p_s is not valid XHTML but standard HTML, where
        # invalid-XML tags like <br> may appear.

        # If, in p_s, you have inserted special "a" tags allowing to inject
        # specific content, like parts of external Python source code files or
        # PX, set parameter p_inject to True. In that case, everytime such a
        # link will be encountered, it will replaced with the appropriate
        # content. In the case of an external Python source code file, POD will
        # retrieve it via a HTTP request and incorporate the part (class,
        # method..) specified in the link's "title" attribute, into the
        # resulting chunk of ODF code. If a PX is specified, it will be called
        # with the current PX context, and the result will be injected where the
        # link has been defined. For more information about this functionality,
        # check Appy's Rich field in class appy.model.fields.rich.c_Rich.

        # If you set p_unwrap being True, if the resulting ODF chunk is
        # completely included in a single paragraph (<text:p>), this paragraph's
        # start and end tags will be removed from the result (keeping only its
        # content). Avoid using this. p_unwrap should only be used internally by
        # POD itself.

        stylesMapping = self.stylesManager.checkStylesMapping(stylesMapping)
        if html is None: html = self.html
        return Xhtml2OdtConverter(s, self.stylesManager, stylesMapping,
                                  keepWithNext, keepImagesRatio, imagesMaxWidth,
                                  self, html, inject, unwrap).run()

    def renderText(self, s, prefix=None, tags=None, firstCss=None,
                   otherCss=None, lastCss=None, stylesMapping={}):
        '''Renders the pure text p_s in the ODF result. For every carriage
           return character present in p_s, a new paragraph is created in the
           ODF result.'''

        # p_prefix, if passed, must be a string that will be inserted before
        # p_s's first line, followed by a tab.

        # p_tags can be a sequence of XHTML single-letter tags (ie, "bu" for
        # "bold" + "underline"). If passed, formatting as defined by these
        # letters will be applied to the whole p_s (the p_prefix excepted).

        # You may go further by applying CSS classes to the paragraphs produced
        # by p_s's conversion to ODF. For every such class, if a style having
        # the same name is found on the ODF template, it will be applied to the
        # corresponding paragraph. The CSS class:

        # ----------------------------------------------------------------------
        # defined in      | will be
        # attribute ...   | applied to ...
        # ----------------------------------------------------------------------
        # p_firstCss      | the ODF result's first paragraph;
        # p_lastCss       | the ODF result's last paragraph;
        # p_otherCss      | any other paragraph in between.
        # ----------------------------------------------------------------------

        # Moreover, because, internally, p_s is converted to a chunk of XHTML
        # code, and passed to method m_renderXhtml, a p_stylesMapping can be
        # passed to m_renderText, exactly as you would do for m_renderXhtml.
        # ~~~
        # Determine the prefix to insert before the first line
        prefix = '%s<tab/>' % prefix if prefix else ''
        # Split p_s into lines of text
        r = s.split('\n')
        if tags:
            # Prepare the opening and closing sub-tags
            opening = ''.join(['<%s>' % tag for tag in tags])
            closing = ''
            j = len(tags) - 1
            while j >= 0:
                closing += '</%s>' % tags[j]
                j -= 1
        else:
            opening = closing = ''
        i = length = len(r) - 1
        pre = ''
        while i >= 0:
            # Get the CSS class to apply
            if i == 0:
                css = firstCss
                pre = prefix
            elif i == length:
                css = lastCss
            else:
                css = otherCss
            css = ' class="%s"' % css if css else ''
            r[i] = '<p%s>%s%s%s%s</p>' % (css, pre, opening, r[i], closing)
            i -= 1
        return self.renderXhtml(''.join(r), stylesMapping=stylesMapping)

    # Supported image formats. "image" represents any format
    imageFormats = ('png', 'jpeg', 'jpg', 'gif', 'svg', 'image')
    ooFormats = ('odt',)
    convertibleFormats = FILE_TYPES.keys()
    # Constant indicating if a renderer must inherit a value from is parent
    # renderer.
    INHERIT = 1

    def importDocument(self, content=None, at=None, format=None,
      anchor='as-char', wrapInPara=True, size=None, sizeUnit='cm',
      maxWidth='page', style=None, keepRatio=True, pageBreakBefore=False,
      pageBreakAfter=False, convertOptions=None):
        '''Implements the POD statement "do... from document"'''

        # If p_at is not None, it represents a path or url allowing to find the
        # document. It can be a string or a pathlib.Path instance. If p_at is
        # None, the content of the document is supposed to be in binary format
        # in p_content. The document p_format may be: odt or any format in
        # imageFormats.

        # p_anchor, p_wrapInPara, p_size, p_sizeUnit, p_style and p_keepRatio
        # are only relevant for images:
        # * p_anchor defines the way the image is anchored into the document;
        #       Valid values are 'page', 'paragraph', 'char' and 'as-char';
        # * p_wrapInPara, if true, wraps the resulting 'image' tag into a 'p'
        #       tag;
        # * p_size, if specified, is a tuple of float or integers (width,
        #       height) expressing size in p_sizeUnit (see below). If not
        #       specified, size will be computed from image info;
        # * p_sizeUnit is the unit for p_size elements, it can be "cm"
        #       (centimeters), "px" (pixels) or "pc" (percentage). Percentages,
        #       in p_size, must be expressed as integers from 1 to 100.
        # * if p_maxWidth is specified (as a float value representing cm),
        #       image's width will not be greater than it. If p_maxWidth is
        #       "page" (the default), p_maxWidth will be computed as the width
        #       of the main document's page style, margins excluded.
        # * if p_style is given, it is a appy.shared.css.CssStyles instance,
        #       containing CSS attributes. If "width" and "heigth" attributes
        #       are found there, they will override p_size and p_sizeUnit.
        # * If p_keepRatio is True, the image width/height ratio will be kept
        #       when p_size is specified.

        # p_pageBreakBefore and p_pageBreakAfter are only relevant for importing
        # external odt documents, and allows to insert a page break before/after
        # the inserted document. More precisely, each of these parameters can
        # have values:
        # * True     insert a page break;
        # * False    do no insert a page break;
        # moreover, p_pageBreakAfter can have this additional value:
        # * 'duplex' insert 2 page breaks if the sub-document has an odd number
        #            of pages, 1 else (useful for duplex printing).

        # If p_convertOptions are given (for images only), imagemagick will be
        # called with these options to perform some transformation on the image.
        # For example, if you specify
        #                    convertOptions="-rotate 90"

        # pod will perform this command before importing the file into the
        # result:
        #               convert your.image -rotate 90 your.image

        # You can also specify a function in convertOptions. This function will
        # receive a single arg, "image", an instance of
        # appy.pod.doc_importers.Image giving some characteristics of the image
        # to convert, like image.width and image.height in pixels (integers). If
        # your function does not return a string containing the convert options,
        # no conversion will occur.

        # ~~~
        # Is there something to import ?
        if not content and not at: raise PodError(DOC_KO)
        # Convert p_at to a string if it is not the case
        at = str(at) if isinstance(at, Path) else at
        # Guess document format
        if not format:
            # It should be deduced from p_at
            if not at:
                raise PodError(DOC_FMT_KO)
            format = os.path.splitext(at)[1][1:]
        else:
            # If format is a mimeType, convert it to an extension
            if format in utils.mimeTypesExts:
                format = utils.mimeTypesExts[format]
        isImage = isOdt = False
        importer = None
        if format in self.ooFormats:
            importer = importers.OdtImporter
            self.forceOoCall = True
            isOdt = True
        elif (format in self.imageFormats) or not format:
            # If the format can't be guessed, we suppose it is an image
            importer = importers.ImageImporter
            isImage = True
        elif format == 'pdf':
            importer = importers.PdfImporter
        elif format in self.convertibleFormats:
            importer = importers.ConvertImporter
        else:
            raise PodError(DOC_FMT_NS % format)
        imp = importer(content, at, format, self)
        # Initialise image-specific parameters
        if isImage:
            imp.init(anchor, wrapInPara, size, sizeUnit, maxWidth, style,
                     keepRatio, convertOptions)
        elif isOdt: imp.init(pageBreakBefore, pageBreakAfter)
        return imp.run()

    def importImage(self, content=None, at=None, format=None, size=None,
                    sizeUnit='cm', maxWidth='page', style=None, keepRatio=True,
                    convertOptions=None):
        '''While m_importDocument allows to import a document or image via a
           "do ... from document" statement, method m_importImage allows to
           import an image anchored as a char via a POD expression
           ":image(...)", having almost the same parameters.'''

        # Compared to the "document" function, and due to the specific nature of
        # the "image" function, 2 parameters are missing:
        # (1) "anchor" is forced to be "as-char";
        # (2) "wrapInPara" is forced to False.
        return self.importDocument(content=content, at=at, format=format,
          wrapInPara=False, size=size, sizeUnit=sizeUnit, maxWidth=maxWidth,
          style=style, keepRatio=keepRatio, convertOptions=convertOptions)

    def getResolvedNamespaces(self):
        '''Gets a context where mainly used namespaces have been resolved'''
        env = self.stylesParser.env
        return {'text': env.ns(env.NS_TEXT), 'style': env.ns(env.NS_STYLE)}

    def importPod(self, content=None, at=None, format='odt', context=None,
                  pageBreakBefore=False, pageBreakAfter=False,
                  managePageStyles='rename', resolveFields=False,
                  forceOoCall=INHERIT):
        '''Implements the POD statement "do... from pod"'''

        # Similar to m_importDocument, but allows to import the result of
        # executing the POD template whose absolute path is specified in p_at
        # (or, but deprecated, whose binary content is passed in p_content, with
        # this p_format) and include it in the POD result.

        # Renaming page styles for the sub-POD (p_managePageStyles being
        # "rename") ensures there is no name clash between page styles (and tied
        # elements such as headers and footers) coming from several sub-PODs or
        # with styles defined at the master document level. This takes some
        # processing, so you can set it to None if you are sure you do not need
        # it.

        # p_resolveFields has the same meaning as the homonym parameter on the
        # Renderer.

        # By default, if p_forceOoCall is True for p_self, a sub-renderer ran by
        # a c_PodImporter will inherit from this attribute, excepted if
        # parameter p_forceOoCall is different from INHERIT.
        # ~~~
        # Is there a pod template defined ?
        if not content and not at: raise PodError(DOC_KO)
        imp = importers.PodImporter(content, at, format, self)
        self.forceOoCall = True
        # Define the context to use: either the current context of the current
        # POD renderer, or p_context if given.
        context = context or self.contentParser.env.context
        imp.init(context, pageBreakBefore, pageBreakAfter, managePageStyles,
                 resolveFields, forceOoCall)
        return imp.run()

    def importCell(self, content, style='Default'):
        '''Creates a chunk of ODF code ready to be dumped as a table cell
           containing this p_content and styled with this p_style.'''
        return '<table:table-cell table:style-name="%s">' \
               '<text:p>%s</text:p></table:table-cell>' % (style, content)

    def drawShape(self, name, type='rect', stroke='none', strokeWidth='0cm',
                  strokeColor='#666666', fill='solid', fillColor='#666666',
                  anchor='char', width='1.0cm', height='1.0cm',
                  deltaX='0.0cm', deltaY='0.0cm', target='styles'):
        '''Renders a shape within a POD result'''
        # Generate a style for this shape and add it among dynamic styles
        style = self.stylesManager.stylesGenerator.addGraphicalStyle(target,
          stroke=stroke, strokeWidth=strokeWidth, strokeColor=strokeColor,
          fill=fill, fillColor=fillColor)
        # Return the ODF code representing the shape
        return '<draw:%s text:anchor-type="%s" draw:z-index="1" ' \
               'draw:name="%s" draw:style-name="%s" ' \
               'draw:text-style-name="MP1" svg:width="%s" svg:height="%s"' \
               ' svg:x="%s" svg:y="%s"><text:p/></draw:%s>' % \
               (type, anchor, name, style.name, width, height, deltaX, deltaY,
                type)

    def _insertBreak(self, type):
        '''Inserts a page or column break into the result'''
        name = 'podPageBreakAfter' if type == 'page' else 'podColumnBreak'
        return '<text:p text:style-name="%s"></text:p>' % name

    def insertPageBreak(self): return self._insertBreak('page')
    def insertColumnBreak(self): return self._insertBreak('column')

    def prepareFolders(self):
        '''Ensure p_self.result is correct and create, when relevant, the temp
           folder for preparing it.'''
        # Ensure p_self.result is an absolute path
        self.result = os.path.abspath(self.result)
        # Raise an error if the result already exists and we can't overwrite it
        if os.path.isdir(self.result):
            raise PodError(RES_FOLDER % self.result)
        exists = os.path.isfile(self.result)
        if not self.overwriteExisting and exists:
            raise PodError(RES_EXISTS % self.result)
        # Remove the result if it exists
        if exists: os.remove(self.result)
        # Create a temp folder for storing temporary files
        self.tempFolder = '%s.%f' % (self.result, time.time())
        try:
            os.mkdir(self.tempFolder)
        except OSError as oe:
            raise PodError(TEMP_W_KO % (self.result, oe))
        # The "unzip" folder is a sub-folder, within self.tempFolder, where
        # p_self.template will be unzipped.
        self.unzipFolder = os.path.join(self.tempFolder, 'unzip')
        os.mkdir(self.unzipFolder)

    def patchMetadata(self):
        '''Declares, in META-INF/manifest.xml, images or files included via the
           "do... from document" statements if any, and patch meta.xml (field
           "title").'''
        # Patch META-INF/manifest.xml
        j = os.path.join
        if self.fileNames:
            toInsert = ''
            for fileName in self.fileNames.keys():
                mimeType = mimetypes.guess_type(fileName)[0]
                toInsert += ' <manifest:file-entry manifest:media-type="%s" ' \
                            'manifest:full-path="%s"/>\n' % (mimeType, fileName)
            manifestName = j(self.unzipFolder, 'META-INF', 'manifest.xml')
            # Read the the content of this file, if not already in
            # self.manifestXml.
            if not self.manifestXml:
                f = open(manifestName, **enc)
                self.manifestXml = f.read()
                f.close()
            hook = '</manifest:manifest>'
            content = self.manifestXml.replace(hook, toInsert + hook)
            # Write the new manifest content
            f = open(manifestName, 'w', **enc)
            f.write(content)
            f.close()
        # Patch meta.xml
        metadata = self.metadata
        if metadata:
            metaName = j(self.unzipFolder, 'meta.xml')
            # Read the content of this file, if not already in self.metaXml
            if not self.metaXml:
                f = open(metaName, **enc)
                self.metaXml = f.read()
                f.close()
            # Remove the existing title, if it exists
            content = self.metaRex.sub('', self.metaXml)
            # Add a new title, based on the result name
            if isinstance(metadata, str):
                title = metadata
            else:
                title = os.path.splitext(os.path.basename(self.result))[0]
            hook = self.metaHook
            title = '<dc:title>%s</dc:title>%s' % (title, hook)
            content = content.replace(hook, title)
            f = open(metaName, 'w', **enc)
            f.write(content)
            f.close()

    # Public interface
    def run(self):
        '''Renders the result'''
        try:
            # Remember which parser is running
            self.currentParser = self.contentParser
            # Create the resulting content.xml
            self.currentParser.parse(self.contentXml)
            self.currentParser = self.stylesParser
            # Create the resulting styles.xml
            self.currentParser.parse(self.stylesXml)
            # Patch metadata
            self.patchMetadata()
            # Re-zip the result
            self.finalize()
        finally:
            try:
                if self.deleteTempFolder:
                    FolderDeleter.delete(self.tempFolder)
            except PermissionError:
                # The temp folder could not be deleted
                pass

    def getStyles(self):
        '''Returns a dict of the styles that are defined into the template'''
        return self.stylesManager.styles

    def setStylesMapping(self, stylesMapping):
        '''Establishes a correspondence between, on one hand, CSS styles or
           XHTML tags that will be found inside XHTML content given to POD,
           and, on the other hand, ODT styles found into the template.'''
        try:
            manager = self.stylesManager
            # Initialise the styles mapping when relevant
            ocw = self.optimalColumnWidths
            dis = self.distributeColumns
            if ocw or dis:
                sm.TableProperties.initStylesMapping(stylesMapping, ocw, dis)
            manager.stylesMapping = manager.checkStylesMapping(stylesMapping)
        except PodError as po:
            self.contentParser.env.currentBuffer.content.close()
            self.stylesParser.env.currentBuffer.content.close()
            if os.path.exists(self.tempFolder):
                FolderDeleter.delete(self.tempFolder)
            raise po

    def callLibreOffice(self, resultName, outputName=None, format=None):
        '''Call LibreOffice in server mode to convert or update the result'''
        # Here are the valid combinations for parameter's values.
        # ----------------------------------------------------------------------
        # no outputName | The converter will produce a result whose name is
        #   no format   | p_resultName whose file extension has been replaced
        #               | with the one specified in self.resultType.
        # ----------------------------------------------------------------------
        # no outputName | Idem, but the output format is forced to p_format
        #    format     | instead of using self.resultType.
        # ----------------------------------------------------------------------
        #  outputName   | The converter will produce a result whose name is
        #  no format    | p_outputName. The output format will depend on
        #               | p_outputName's file extension.
        # ----------------------------------------------------------------------
        loOutput = ''
        # Determine the "result" to ask to LibreOffice
        if outputName:
            result = outputName
            resultType = os.path.splitext(result)[-1][1:]
        else:
            result = format or self.resultType
            resultType = result
        # Get CSV options if we must produce a CSV file
        if resultType == 'csv':
            csvOptions = (self.csvOptions or CsvOptions.default).asString()
        else:
            csvOptions = None
        try:
            try:
                from appy.pod.converter import Converter
                try:
                    Converter(resultName, result, self.ooServer, self.ooPort,
                      self.stylesTemplate, self.optimalColumnWidths,
                      self.distributeColumns, self.script, self.resolveFields,
                      self.pdfOptions, csvOptions, self.ppp, self.stream,
                      self.pageStart).run()
                except Converter.Error as ce:
                    raise PodError(CONV_ERROR % str(ce))
            except ImportError:
                # I do not have UNO. So try to launch a UNO-enabled Python
                # interpreter which should be in self.pythonWithUnoPath.
                pyPath = self.pythonWithUnoPath
                if not pyPath:
                    raise PodError(NO_PY_PATH % resultType)
                if not os.path.isfile(pyPath):
                    raise PodError(PY_PATH_KO % pyPath)
                convScript = '%s/converter.py' % \
                             os.path.dirname(appy.pod.__file__)
                cmd = [pyPath, convScript, resultName, result,
                       '-e', self.ooServer, '-p', str(self.ooPort)]
                add = cmd.append
                if self.stylesTemplate:
                    add('-t'); add(self.stylesTemplate)
                if self.optimalColumnWidths:
                    add('-o'); add(str(self.optimalColumnWidths))
                if self.distributeColumns:
                    add('-d'); add(str(self.distributeColumns))
                if self.script:
                    add('-s'); add(self.script)
                if self.resolveFields:
                    add('-r'); add(self.resolveFields)
                if self.pdfOptions:
                    add('-f'); add(self.pdfOptions)
                if csvOptions:
                    add('-i'); add(csvOptions)
                if self.ppp: add('-c')
                if self.stream != 'auto':
                    add('-a'); add(str(self.stream))
                if self.pageStart > 1:
                    add('-g'); add(str(self.pageStart))
                out, loOutput = utils.executeCommand(cmd)
        except PodError as pe:
            # When trying to call LO in server mode for producing ODT or ODS
            # (=forceOoCall=True), if an error occurs we have nevertheless
            # an ODT or ODS to return to the user. So we produce a warning
            # instead of raising an error.
            if (resultType in self.templateTypes) and self.forceOoCall:
                print(WARN_OD_KO % str(pe))
            else:
                raise pe
        return loOutput

    def getTemplateType(self, template):
        '''Identifies the type of this POD p_template (ods or odt). If
           p_template is a string, it is a file name: simply get its extension.
           Else, it is a binary file in a StringIO instance: seek the MIME type
           from the first bytes.'''
        if isinstance(template, str):
            r = os.path.splitext(template)[1][1:]
        else:
            # A StringIO instance
            template.seek(0)
            sbytes = template.read(90)
            sbytes = sbytes[sbytes.index('mimetype')+8:]
            r = 'ods' if sbytes.startswith(utils.mimeTypes['ods']) else 'odt'
        return r

    def finalize(self):
        '''Re-zip the result and potentially call LibreOffice if target format
           is not among self.templateTypes or if forceOoCall is True.'''
        j = os.path.join
        # If an action regarding page styles must be performed, get and modify
        # them accordingly.
        pageStyles = None
        mps = self.managePageStyles
        if mps is not None:
            pageStyles = self.stylesManager.pageStyles.init(mps)
        # Patch styles.xml and content.xml
        dynamic = self.stylesManager.dynamicStyles
        for name in ('styles', 'content'):
            # Copy the [content|styles].xml file from the temp to the zip folder
            fn = '%s.xml' % name
            shutil.copy(j(self.tempFolder, fn), j(self.unzipFolder, fn))
            # Get the file content
            fn = os.path.join(self.unzipFolder, fn)
            f = open(fn, **enc)
            content = f.read()
            f.close()
            # Inject self.fonts, when present, in styles.xml
            isStylesXml = name == 'styles'
            if isStylesXml and self.fonts:
                content = sm.FontsInjector(self.fonts).injectIn(content)
            # Inject dynamic styles
            if isStylesXml: content = dynamic.injectIn('styles_base', content)
            content = dynamic.injectIn(name, content)
            # Rename the page styles
            if pageStyles:
                content = pageStyles.renameIn(name, content)
            # Write the updated content to the file
            f = open(fn, 'w', **enc)
            f.write(content)
            f.close()
        # Call the user-defined "finalize" function(s) when present
        if self.finalizeFunction:
            try:
                for fun in self.finalizeFunction: fun(self.unzipFolder, self)
            except Exception as e:
                print(WARN_FIN_KO % str(e))
        # Re-zip the result, first as an OpenDocument file of the same type as
        # the POD template (odt, ods...)
        resultName = os.path.join(self.tempFolder,
                                  'result.%s' % self.templateType)
        resultType = self.resultType
        zip(resultName, self.unzipFolder, odf=True)
        if (resultType in self.templateTypes) and not self.forceOoCall:
            # Simply move the ODT result to the result
            os.rename(resultName, self.result)
        else:
            if resultType not in FILE_TYPES:
                raise PodError(R_TYPE_KO % (self.result, FILE_TYPES.keys()))
            # Call LibreOffice to perform the conversion or document update
            output = self.callLibreOffice(resultName, self.result)
            # I (should) have the result in self.result
            if not os.path.exists(self.result):
                if resultType in self.templateTypes:
                    # In this case LO in server mode could not be called (to
                    # update indexes, sections, etc) but I can still return the
                    # "raw" pod result that exists in "resultName".
                    os.rename(resultName, self.result)
                else:
                    raise PodError(CONV_ERROR % output)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This module contains all visitors for the tree of tags defined in tags.py
   and build by parser.py.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.pod.xhtml import tags

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Visitor:
    '''Abstract base class for all visitors'''

    def visit(self, env):
        '''Visits the tree of tags that has been build in the p_env(ironment)'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TablesVisitor(Visitor):
    '''Abstract visitor that walks through all tables'''

    def visit(self, env):
        # Do nothing if no table was found
        if 'Table' not in env.tags: return
        # Walk every table
        r = False
        for table in env.tags['Table']:
            updated = self.visitTable(table)
            r = r or updated
        return updated

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TablesNormalizer(TablesVisitor):
    '''Ensure all tables rows are under a tag "thead" or "tbody"'''

    def isHeaderRow(self, row):
        '''A row is considered a header row if all cells are "th" cells'''
        for cell in row.iterChildren(types='Cell'):
            if cell.name == 'td': return
        return True # Implicitly, all rows are "th" rows if we are here

    def visitTable(self, table):
        '''Visiting a table = normalizing it'''
        # Store here tags "thead" and "tbody"
        thead = tbody = None
        # Store here "root" rows, ie, those encountered directly under the
        # "table" tag (not being under tag "thead" nor "tbody").
        rootRows = []
        # Walk table sub-tags
        for tag in table.iterChildren():
            if tag.className == 'Header':
                thead = tag
            elif tag.className == 'Body':
                tbody = tag
            elif tag.className == 'Row':
                rootRows.append(tag)
        # Set all root rows within thead or tbody
        for row in rootRows:
            if self.isHeaderRow(row):
                # This is a header row. Create the "thead" tag when inexistent.
                if not thead:
                    thead = tags.Header('thead', None, parent=table)
                row.setParent(thead)
            else:
                # This is a body row. Create the "tbody" tag when inexistent.
                if not tbody:
                    tbody = tags.Body('tbody', None, parent=table)
                row.setParent(tbody)
        # We have currently no interest of reporting if at least one table has
        # been updated or not.
        return True

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TablesOptimizer(TablesVisitor):
    '''Removes most style information about tables'''

    # Attributes to remove on tag "table" and sub-tags
    undesired = ('width', 'height')
    undesiredAttributes = {
      'table': undesired + ('style',),
      'tr'   : undesired + ('style',),
      'th'   : undesired,
      'td'   : undesired
    }

    # CSS properties to keep within "style" attributes
    desiredProperties = ('text-align',)

    def removeAttributes(self, tag):
        '''Remove any undesired attribute on this p_tag'''
        # Returns true if at least one removal has been performed
        r = False
        # Remove undesired attributes on the p_table
        attrs = tag.attrs
        if not attrs: return
        # Walk undesired attributes
        for name in self.undesiredAttributes[tag.name]:
            if hasattr(attrs, name):
                delattr(attrs, name)
                r = True
        return r

    def removeProperties(self, tag):
        '''Remove ay CSS property from p_tag's "style" attribute not being in
           self.desiredProperties.'''
        props = tag.attrs.style.strip(' ;').split(';')
        if not props: return
        # Walk CSS properties in reverse order and remove the undesired ones
        keep = self.desiredProperties
        i = len(props) - 1
        updated = False
        while i >= 0:
            try:
                name, value = props[i].split(':', 1)
                name = name.strip()
            except ValueError:
                # The value is wrong
                del props[i]
                updated = True
            # Remove it when appropriate
            if name not in keep:
                del props[i]
                updated = True
            i -= 1
        # Set the updated style info on p_tag, if it has been updated
        if updated:
            # All proprerties may have been removed
            if not props:
                delattr(tag.attrs, 'style')
            else:
                tag.attrs.style = ';'.join(props)
        return updated

    def visitRow(self, row):
        '''Optimize this table p_row'''
        updated = self.removeAttributes(row)
        # Walk cells
        for cell in row.iterChildren(types='Cell'):
            # Remove undesired attributes
            r = self.removeAttributes(cell)
            updated = updated or r
            # Remove undesired CSS properties
            attrs = cell.attrs
            if attrs and hasattr(attrs, 'style'):
                r = self.removeProperties(cell)
                updated = updated or r
        return updated

    def visitTable(self, table):
        '''Visiting a table = optimizing it'''
        # Remove undesired attributes on the p_table
        updated = self.removeAttributes(table)
        # Ensure the table as simple 1px border
        attrs = table.attrs
        if not attrs or not hasattr(attrs, 'border') or (attrs.border != '1'):
            table.addAttribute('border', 1)
            updated = True
        # Remove undesired attributes on sub-tags: tr, th and td. Find and walk
        # all table rows.
        for tag in table.iterChildren(types=('Row', 'Header', 'Body')):
            if tag.name == 'tr':
                r = self.visitRow(tag)
                updated = updated or r
            elif tag.name in ('thead', 'tbody'):
                for row in tag.iterChildren(types='Row'):
                    r = self.visitRow(row)
                    updated = updated or r
        return updated
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class KeepWithNext(Visitor):
    # CSS classes with "keep with next" functionality, by tag type. By default,
    # class "ParaKWN" is applied.
    cssClasses = {'li': 'podItemKeepWithNext'}
    # We will only walk tags directly under the root tag, having one of these
    # types.
    rootTagTypes = ('Para', 'List', 'Table')

    def __init__(self, chars):
        # The number of chars that must be kept together
        self.charsToKeep = chars
        # The number of chars walked and counted so far
        self.charsWalked = 0
        # If we must split a table due to our "keep-with-next" constraints, we
        # store here the table row of the table to split; this row will be the
        # first of the second table.
        self.splitRow = None

    def charsReached(self):
        '''Have we reached self.charsToKeep ?'''
        return self.charsWalked >= self.charsToKeep

    def splitTable(self):
        '''Splits the table containing p_self.splitRow into 2 separate tables in
           order to enforce "keep-with-next" constraints.'''
        # Let's call the table to split "table" and the splitted tables
        # "table1" and "table2". Instead of performing a real split, where
        # "table1" will contain the first "table" rows, including the one
        # preceding p_self.splitRow, and "table2" will contain the last "table"
        # rows, from p_self.splitRow to the end, we will duplicate "table":
        # "table1" and "table2" will contain all "table" rows. Indeed, xhtml2odt
        # and LibreOffice will potentially compute cell's widths, and if these
        # computations are performed on "table1" and "table2" respectively
        # containing their corresponding rows, they may produce different
        # results on "table1" and on "table2". That is not acceptable. This is
        # why we keep all rows duplicated in both tables, but we set a
        # "keeprows" attribute that will be processed by the POD's
        # post-processor.
        # Get the first table's tags of interest
        tbody1 = self.splitRow.parent
        table1 = tbody1.parent
        # Count the number of rows within table1's header
        header1 = table1.getChild('Header')
        if header1:
            headerRows = header1.countChildren('Row')
        else:
            headerRows = 0
        # Create the second table by cloning the first one
        table2 = table1.clone()
        tbody2 = tags.Body('tbody', parent=table2)
        # Ensure "table2" will be kept with the next paragraphs
        table2.addCss('TableKWN')
        # Copy all rows from "table1" to "table2". Here, we cheat: instead of
        # copying rows, we copy them by reference. The rows will have "tbody1"
        # as parent, but currently this has no importance when dumping the tree
        # of tags as XHTML.
        tbody2.children = tbody1.children
        # Set attributes "keeprows" on "table1" and "table2"
        i = tbody1.children.index(self.splitRow)
        table1.addAttribute('keeprows', ':%d' % (headerRows+i))
        table1.addAttribute('style', 'margin-bottom:0px')
        table2.addAttribute('keeprows', '%d:' % i)
        table2.addAttribute('style', 'margin-top:0px')
        # Insert table2 just after table1
        parent = table1.parent
        table2.setParent(parent, at=parent.children.index(table1)+1)

    def visitPara(self, para):
        '''Visiting a p_para(graph) = applying the correct "keep-with-next"
           class on it.'''
        para.addCss(self.cssClasses.get(para.name) or 'ParaKWN')
        # Update the number of walked chars
        self.charsWalked += para.getContentLength()

    def visitList(self, list):
        '''Visiting a list = visiting its items in reverse order'''
        # "li" sub-tags are Para instances
        for child in list.iterChildren(types='Para', reverse=True):
            self.visitPara(child)
            if self.charsReached():
                return

    def visitRow(self, row):
        '''We visit a row only for computing its content length'''
        self.charsWalked += row.getContentLength()

    def visitTable(self, table):
        '''Visiting a table = visiting its rows in reverse order.

           If we reach p_self.charsToKeep at a given row R, we must split the
           table in 2 tables: the first one containing all rows preceding R, and
           the second one containing row R and subsequent rows.
        '''
        # Get the "tbody" tag
        tbody = table.getChild('Body')
        if not tbody: return
        # Walk tbody's rows in reverse order
        for row in tbody.iterChildren(types='Row', reverse=True):
            self.visitRow(row)
            if self.charsReached():
                # This row and subsequent ones must be splitted in a second
                # table, excepted if this row is the first row of data in its
                # table. Indeed, we will not take the risk of having table
                # headers left alone on a separate page.
                if row != tbody.getChild('Row'):
                    self.splitRow = row
                break

    def visit(self, env):
        '''We will determine the set of root paragraphs corresponding to the
           last "self.chars" chars and apply, on each one, a special class that
           will ensure it will be kept with the next one on the same page.'''
        # Walk tags directly under the root tag, in reverse order
        for child in env.r.iterChildren(types=self.rootTagTypes, reverse=True):
            # Visit the child
            getattr(self, 'visit%s' % child.className)(child)
            # Stop if we have reached the number of chars to keep together
            if self.charsReached():
                break
        # If we have reached chars in the middle of a table, we must split it
        if self.splitRow: self.splitTable()
        return True

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Cleaner(Visitor):
    '''Removes empty paragraphs at the end of the text and/or between non empty
       paragraphs.'''

    def __init__(self, removeTrailingParas=True):
        # If p_removeTrailingParas is...
        # ----------------------------------------------------------------------
        # True  | empty paragraphs situated at the end of the text will be
        #       | removed. Empty paragraphs found between 2 non empty paragraphs
        #       | will be left untouched.
        # ----------------------------------------------------------------------
        # False | empty paragraphs found between 2 non empty paragraphs will be
        #       | removed. Those situated at the end of the text will be left
        #       | untouched.
        # ----------------------------------------------------------------------
        # None  | Nothing is done regarding empty paragraphs.
        # ----------------------------------------------------------------------
        self.removeTrailingParas = removeTrailingParas

    def isEmptyPara(self, child):
        '''Does p_child represent an empty paragraph or something we can ignore
           (like a Content instance) ?'''
        return not child.isTag or \
               (isinstance(child, tags.Para) and child.isEmpty())

    def removeEndingParas(self, root):
        '''Remove trailing empty paragraphs'''
        updated = False
        for child in root.iterChildren(reverse=True):
            # Content at this root level is garbage
            if self.isEmptyPara(child):
                root.children.remove(child)
                updated = True
            else:
                # Stop here, we have met something non empty
                break
        return updated

    def removeEmptyParas(self, root):
        '''Remove empty paragraphs being inserted between 2 non empty
           paragraphs.'''
        updated = False
        # Browse tags in reverse order. Do not touch the end of the content
        # until non empty content is found (removing stuff in this part is done
        # via m_removeEndingParas, if enabled). Then, remove any empty paragraph
        # found.
        inBottom = True
        # Remember the previously encountered child (=the one following the
        # current child, because we walk then in reverse order).
        previous = None
        for child in root.iterChildren(reverse=True):
            # Ignore empty content at the bottom of the text
            if self.isEmptyPara(child):
                if inBottom:
                    continue
                # Remove this empty para and the possible carriage return
                # that follows (<p></p>\n).
                if child.isTag:
                    root.children.remove(child)
                    if previous and not previous.isTag:
                        root.children.remove(previous)
                    updated = True
            else:
                # We have met something non empty
                inBottom = False
            previous = child
        return updated

    def visit(self, env):
        '''Removes the appropriate paragraphs'''
        remove = self.removeTrailingParas
        if remove is None: return
        method = remove and 'removeEndingParas' or 'removeEmptyParas'
        return getattr(self, method)(env.r)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Data structures for lazily accessing elements stored in sub-structures'''

# Code inspired by Zope's ZTUtil/Lazy.py

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
OVERRIDE_ERROR = 'Lazy subclasses must implement __getitem__.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Lazy:
    '''Abstract base class for any lazy data structure'''

    def __init__(self, seq, length=None):
        # The sequence of elements for which lazy access will be enabled
        self.seq = seq
        # The total length of p_seq's sub-elements, if known
        self.length = len(seq) if length is None else length

    def __repr__(self):
        '''p_self's string representation'''
        return repr(list(self))

    def __len__(self): return self.length

    def __getitem__(self, index):
        if isinstance(index, slice):
            r = []
            start, stop, step = index.indices(len(self))
            for i in range(start, stop, step):
                try:
                    r.append(self[i])
                except IndexError:
                    return r
            return r
        # The single index lookup is implemented in the subclasses
        raise NotImplementedError(OVERRIDE_ERROR)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class LazyValues(Lazy):
    '''Represents a list of (key, value) tuples, with lazy access to the "value"
       parts.'''

    def __getitem__(self, index):
        '''Accesses the "value" part of tuple element at p_index'''
        return self.__class__(self.seq[index]) if isinstance(index, slice) \
                                               else self.seq[index][1]

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class LazyCat(Lazy):
    '''Represents a concatenation of sub-lists whose elements are lazily
       accessed as if a single global list was existing.'''

    def __init__(self, seq, length=None):
        Lazy.__init__(self, seq, length)
        # The global list that will be built on-demand, every time an element
        # needs to be accessed. When p_self.data will be completely built,
        # p_self.seq will be set to None.
        self.data = []
        # The index of the last walked element within p_seq
        self.sindex = 0
        # The index of the last walked sub-element within p_seq[p_self.sindex]
        self.eindex = -1

    def __len__(self):
        '''Get p_self's length, ie, the sum of lengths of sub-sequences'''
        if self.length is not None: return self.length
        r = 0
        if self.seq is None:
            r = len(self.data)
        else:
            for l in self.seq:
                r += len(l)
        self.length = r
        return r

    def __getitem__(self, index):
        '''Accesses the sub-element at this p_index, as if p_self was a flat
           list.'''
        # Slices are managed by the base method
        if isinstance(index, slice):
            return super().__getitem__(index)
        # If the part of the global list (p_self.data) that includes p_index is
        # already available, get the element from it.
        seq = self.seq
        if seq is None:
            return self.data[index]
        # Manage negative indexes
        i = index
        if i < 0:
            i = len(self) + i
        if i < 0:
            raise IndexError(index)
        # Get the element from p_self.data if possible
        data = self.data
        ind = len(data)
        if i < ind:
            return data[i]
        ind -= 1
        # Complete p_self.data until we can access the element at p_index
        sindex = self.sindex
        try:
            s = seq[sindex]
        except Exception:
            raise IndexError(index)
        eindex = self.eindex
        while i > ind:
            try:
                eindex += 1
                v = s[eindex]
                data.append(v)
                ind += 1
            except IndexError:
                self.sindex = sindex = sindex + 1
                try:
                    s = self.seq[sindex]
                except Exception:
                    # We have completely walked p_self.seq. p_self.data is
                    # complete.
                    self.seq = self.sindex = self.eindex = None
                    raise IndexError(index)
                self.eindex = eindex = -1
        self.eindex = eindex
        return data[i]
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

import os, os.path, sys, time
from optparse import OptionParser
from appy.shared.utils import FolderDeleter, Traceback
from appy.shared.errors import InternalError
from appy.shared.tables import TablesParser
from appy.shared.xml_parser import XmlComparator

# ------------------------------------------------------------------------------
class TesterError(Exception): pass

# TesterError-related constants
WRONG_TEST_PLAN = 'The test plan you specified does not correspond to an ' \
  'existing ODT file.'
_FLAVOUR = 'A flavour represents a test configuration.'
FLAVOURS_NOT_LIST = 'The flavours specified must be a list or tuple of ' \
  'string. ' + _FLAVOUR
FLAVOUR_NOT_STRING = 'Each specified flavour must be a string. ' + _FLAVOUR
WRONG_TEST_FACTORY = 'You must give a test factory that inherits from the ' \
  'abstract "appy.shared.test.TestFactory" class.'
CREATE_TEST_NOT_OVERRIDDEN = 'The appy.shared.test.TestFactory.createTest ' \
  'method must be overridden in your concrete TestFactory.'
MAIN_TABLE_NOT_FOUND = 'No table "TestSuites" found in test plan "%s".'
MAIN_TABLE_MALFORMED = 'The "TestSuites" table must have at least two ' \
  'columns, named "Name" and "Description".'
TEST_SUITE_NOT_FOUND = 'Table "%s.descriptions" and/or "%s.data" were not ' \
  'found.'
TEST_SUITE_MALFORMED = 'Tables "%s.descriptions" and "%s.data" do not have ' \
  'the same length. For each test in "%s.data", You should have one line in ' \
  '"%s.descriptions" describing the test.'
FILE_NOT_FOUND = 'File to compare "%s" was not found.'
WRONG_ARGS = 'You must specify as unique argument the configuration flavour ' \
             'you want, which may be one of %s.'
WRONG_FLAVOUR = 'Wrong flavour "%s". Flavour must be one of %s.'
TEST_AND_SUITE = 'Specify a single test or test suite but not both.'

# InternalError-related constants
TEST_REPORT_SINGLETON_ERROR = 'You can only use the TestReport constructor ' \
  'once. After that you can access the single TestReport instance via the ' \
  'TestReport.instance static member.'

# ------------------------------------------------------------------------------
class TestReport:
    instance = None
    def __init__(self, testReportFileName, verbose, encoding='utf-8'):
        self.encoding = encoding
        if TestReport.instance == None:
            self.report = open(testReportFileName, 'w')
            self.verbose = verbose
            TestReport.instance = self
        else:
            raise InternalError(TEST_REPORT_SINGLETON_ERROR)

    def say(self, msg, force=False):
        if self.verbose or force:
            print(msg)
        if self.encoding:
            self.report.write(msg.encode(self.encoding))
        else:
            self.report.write(msg)
        self.report.write('\n')

    def close(self): self.report.close()

# ------------------------------------------------------------------------------
class Test:
    '''Abstract test class'''
    def __init__(self, testData, testDescription, testFolder,
                 config, flavour, rendererParams):
        self.data = testData
        self.description = testDescription
        self.testFolder = testFolder
        self.tempFolder = None
        self.report = TestReport.instance
        self.errorDump = None
        self.config = config
        self.flavour = flavour
        self.rendererParams = rendererParams

    def compareFiles(self, expected, actual, areXml=False, xmlTagsToIgnore=(),
                     xmlAttrsToIgnore=(), encoding=None):
        '''Compares 2 files. r_ is True if files are different. The differences
        are written in the test report.'''
        for f in expected, actual:
            assert os.path.exists(f), TesterError(FILE_NOT_FOUND % f)
        # Expected result (may be different according to flavour)
        if self.flavour:
            expectedFlavourSpecific = '%s.%s' % (expected, self.flavour)
            if os.path.exists(expectedFlavourSpecific):
                expected = expectedFlavourSpecific
        # Perform the comparison
        comparator = XmlComparator(actual, expected, areXml, xmlTagsToIgnore,
                                   xmlAttrsToIgnore)
        return not comparator.filesAreIdentical(
            report=self.report, encoding=encoding)

    def run(self):
        self.report.say('-' * 79)
        self.report.say('- Test %s.' % self.data['Name'])
        self.report.say('- %s\n' % self.description)
        # Prepare test data
        self.tempFolder = os.path.join(self.testFolder, 'temp')
        if os.path.exists(self.tempFolder):
            FolderDeleter.delete(self.tempFolder)
        os.mkdir(self.tempFolder)
        try:
            self.do()
            self.report.say('Checking result...')
            testFailed = self.checkResult()
        except:
            testFailed = self.onError()
        self.finalize()
        return testFailed

    def do(self):
        '''Concrete part of the test. Must be overridden.'''

    def checkResult(self):
        '''r_ is False if the test succeeded.'''
        return True

    def onError(self):
        '''What must happen when an exception is raised during test
           execution? Returns True if the test failed.'''
        self.errorDump = Traceback.get()
        self.report.say('Exception occurred:')
        self.report.say(self.errorDump)
        return True

    def finalize(self):
        '''Performs sme cleaning actions after test execution.'''
        pass

    def isExpectedError(self, expectedMessage):
        '''An exception was thrown. So check if the actual error message
           (stored in self.errorDump) corresponds to the p_expectedMessage.'''
        res = True
        for line in expectedMessage:
            if (self.errorDump.find(line) == -1):
                res = False
                self.report.say('"%s" not found among error dump.' % line)
                break
        return res

# ------------------------------------------------------------------------------
class TestFactory:
    @staticmethod
    def createTest(testData, testDescription, testFolder, config, flavour,
                   rendererParams):
        '''This method allows you to create tests that are instances of classes
           that you create. Those classes must be children of
           appy.shared.test.Test. m_createTest must return a Test instance and
           is called every time a test definition is encountered in the test
           plan.'''
        raise TesterError(CREATE_TEST_NOT_OVERRIDDEN)

# ------------------------------------------------------------------------------
class Tester:
    def __init__(self, testPlan, flavours, testFactory):
        # Check test plan
        if (not os.path.exists(testPlan)) or (not os.path.isfile(testPlan)) \
           or (not testPlan.endswith('.odt')):
            raise TesterError(WRONG_TEST_PLAN)
        self.testPlan = testPlan
        self.testFolder = os.path.abspath(os.path.dirname(testPlan))
        # Check flavours
        if (not isinstance(flavours,list)) and (not isinstance(flavours,tuple)):
            raise TesterError(FLAVOURS_NOT_LIST)
        for flavour in flavours:
            if not isinstance(flavour, basestring):
                raise TesterError(FLAVOUR_NOT_STRING)
        self.flavours = flavours
        self.flavour = None
        # Check test factory
        if not issubclass(testFactory, TestFactory):
            raise TesterError(WRONG_TEST_FACTORY)
        self.testFactory = testFactory
        self.getOptions()
        self.report = TestReport('%s/Tester.report' % self.testFolder,
                                 self.verbose)
        self.startTime = time.time()
        self.report.say('Parsing ODT file... ')
        self.tables = TablesParser(testPlan).run()
        self.config = None
        ext = ''
        if self.flavour:
            ext = '.%s' % self.flavour
        configTableName = 'Configuration%s' % ext
        if self.tables.has_key(configTableName):
            self.config = self.tables[configTableName].asDict()
        self.tempFolder = os.path.join(self.testFolder, 'temp')
        if os.path.exists(self.tempFolder):
            FolderDeleter.delete(self.tempFolder)
        self.nbOfTests = 0
        self.nbOfSuccesses = 0
        self.nbOfIgnoredTests = 0

    def getRendererParams(self, params):
        '''Returns a dict with additional Renderer params'''
        res = {}
        if not params: return res
        for param in params.split(','):
            name, value = param.split('=')
            res[name] = value
        return res

    def getOptions(self):
        optParser = OptionParser()
        optParser.add_option('-v', '--verbose', action='store_true',
          help='Dumps the whole test report on stdout')
        optParser.add_option('-k', '--keepTemp', action='store_true', help = \
          'Keep the temp folder, in order to be able to copy some results ' \
          'and make them expected results when needed.')
        optParser.add_option('-t', '--test', dest="test", default='',
          metavar="TEST", help='Run only a specified test.', type='string')
        optParser.add_option('-s', '--testSuite', dest="testSuite", default='',
          metavar="TESTSUITE", help='Run only a specified test suite.',
          type='string')
        optParser.add_option('-r', '--rendererParams', dest="rendererParams",
          default='', metavar="RENDERERPARAMS", type='string',
          help='Specific renderer parameters, of the form name1=val1,name2=v2')
        (options, args) = optParser.parse_args()
        if self.flavours:
            if len(args) != 1:
                raise TesterError(WRONG_ARGS % self.flavours)
            self.flavour = args[0]
            if not self.flavour in self.flavours:
                raise TesterError(WRONG_FLAVOUR % (self.flavour, self.flavours))
        self.verbose = options.verbose == True
        self.keepTemp = options.keepTemp == True
        self.rendererParams = self.getRendererParams(options.rendererParams)
        # If the user wants to run a single test
        self.singleTest = options.test
        # If the user wants to run a single test suite
        self.singleSuite = options.testSuite
        if self.singleTest and self.singleSuite:
            raise TesterError(TEST_AND_SUITE)

    def runSuite(self, suite):
        self.report.say('*' * 79)
        self.report.say('* Suite %s.' % suite['Name'])
        self.report.say('* %s\n' % suite['Description'])
        i = -1
        for testData in self.tables['%s.data' % suite['Name']]:
            self.nbOfTests += 1
            i += 1
            testName = testData['Name']
            if testName.startswith('_'):
                self.nbOfIgnoredTests += 1
            elif self.singleTest and (testName != self.singleTest):
                self.nbOfIgnoredTests += 1
            else:
                description = self.tables['%s.descriptions' % \
                                          suite['Name']][i]['Description']
                test = self.testFactory.createTest(
                    testData, description, self.testFolder, self.config,
                    self.flavour, self.rendererParams)
                testFailed = test.run()
                if not self.verbose:
                    sys.stdout.write('.')
                    sys.stdout.flush()
                if testFailed:
                    self.report.say('Test failed.\n')
                else:
                    self.report.say('Test successful.\n')
                    self.nbOfSuccesses += 1
    def run(self):
        assert self.tables.has_key('TestSuites'), \
               TesterError(MAIN_TABLE_NOT_FOUND % self.testPlan)
        for testSuite in self.tables['TestSuites']:
            if (not testSuite.has_key('Name')) or \
               (not testSuite.has_key('Description')):
                raise TesterError(MAIN_TABLE_MALFORMED)
            # Must we ignore this test suite ?
            if testSuite['Name'].startswith('_'):
                tsName = testSuite['Name'][1:]
                tsIgnored = True
            else:
                tsName = testSuite['Name']
                tsIgnored = self.singleSuite and (tsName != self.singleSuite)
            assert self.tables.has_key('%s.descriptions' % tsName) \
                   and self.tables.has_key('%s.data' % tsName), \
                   TesterError(TEST_SUITE_NOT_FOUND % (tsName, tsName))
            assert len(self.tables['%s.descriptions' % tsName]) == \
                   len(self.tables['%s.data' % tsName]), \
                   TesterError(TEST_SUITE_MALFORMED % ((tsName,)*4))
            # The suite can still be ignored if self.singleTest is not among it
            innerTests = self.tables['%s.data' % tsName]
            if self.singleTest:
                inSuite = False
                for test in innerTests:
                    name = test['Name']
                    if name.startswith('_'): name = name[1:]
                    if name == self.singleTest:
                        inSuite = True
                        break
                if not inSuite: tsIgnored = True
            if tsIgnored:
                nbOfIgnoredTests = len(innerTests)
                self.nbOfIgnoredTests += nbOfIgnoredTests
                self.nbOfTests += nbOfIgnoredTests
            else:
                self.runSuite(testSuite)
        self.finalize()
    
    def finalize(self):
        duration = time.time() - self.startTime
        msg = '%d/%d successful test(s) performed in %.2f seconds' % \
              (self.nbOfSuccesses, (self.nbOfTests-self.nbOfIgnoredTests),
               duration)
        if self.nbOfIgnoredTests >0:
            msg += ', but %d ignored test(s) not counted' % \
                   self.nbOfIgnoredTests
        msg += '.'
        self.report.say(msg, force=True)
        self.report.close()
        if not self.keepTemp:
            if os.path.exists(self.tempFolder):
                FolderDeleter.delete(self.tempFolder)
# ------------------------------------------------------------------------------



'''Module managing translations (i18n)'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.model.utils import Object as O

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class FieldTranslations:
    '''Translations, in all supported languages, for a given field'''

    # Place an instance of this class in attribute "translations" of some field
    # if you do not want Appy to generate, for this field, i18n labels in PO
    # files. With class FieldTranslations, you manage translations "by hand",
    # coding them directly in Python or importing them via a custom tool.

    # Map label types to label attributes
    labelAttrs = {'label': 'labels', 'descr': 'descriptions',
                  'help': 'helps', 'value': 'values'}

    def __init__(self, labels=None, descriptions=None, helps=None, values=None,
                 fallback='en'):
        # Attributes "labels", "descriptions" and "helps" store translations,
        # in every supported language, of the corresponding label type
        # (respectively, the main label, the description label or the help
        # label), and are each of the form:
        #
        #             ~O(en=s_translation, fr=s_translation,...)~
        #
        # Translations for the main field label
        self.labels = labels
        # Translations for the field description
        self.descriptions = descriptions
        # Translation for field helps
        self.helps = helps
        # Translations for possible values of the field. Relevant when a fixed
        # number of values is allowed. This must be a dict of the form:
        #
        #       ~{s_value: O(en=s_translation, fr=s_translation,...)}~
        #
        self.values = values
        # Language to use if the required language is not found
        self.fallback = fallback

    def get(self, label, language, mapping=None, value=None):
        '''Returns the translation, in p_language, of the given p_label type.
           A p_mapping may be given.'''
        # Get the set of translations corresponding to the given p_label type
        translations = getattr(self, FieldTranslations.labelAttrs[label])
        # When getting a field value, dig more deeply
        if value is not None: translations = translations[value]
        # Get, within it, the translation corresponding to p_language, or in the
        # fallback language if p_language is not found.
        r = getattr(translations, language, None) or \
            getattr(translations, self.fallback, '')
        # Todo: apply p_mapping when passed
        return r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy import utils
from appy.model.fields.integer import Integer

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SEP_UNALLOWED = 'Char "%s" is not allowed as decimal separator.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Float(Integer):
    # Allowed chars for being used as decimal separators
    allowedDecimalSeps = (',', '.')

    # Precision of the indexed value, in number of decimals
    indexPrecision = 2

    # The name of the index class storing values of this field in the catalog
    indexType = 'FloatIndex'

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None,
      emptyIndexValue=None, searchable=False, readPermission='read',
      writePermission='write', width=5, height=None, maxChars=13, colspan=1,
      master=None, masterValue=None, focus=False, historized=False,
      mapping=None, generateLabel=None, label=None, sdefault=('',''),
      scolspan=1, swidth=None, sheight=None, persist=True, precision=None,
      sep=(',', '.'), tsep=' ', inlineEdit=False, view=None, cell=None,
      edit=None, xml=None, translations=None, readonly=False,
      alignOnEdit='left', autoComplete=True):
        # The precision is the number of decimal digits. This number is used
        # for rendering the float, but the internal float representation is not
        # rounded.
        self.precision = precision
        # The decimal separator can be a tuple if several are allowed, ie
        # ('.', ',')
        self.sep = sep if type(sep) in utils.sequenceTypes else (sep,)
        # Check that the separator(s) are among allowed decimal separators
        for sep in self.sep:
            if sep not in Float.allowedDecimalSeps:
                raise Exception(SEP_UNALLOWED % sep)
        self.tsep = tsep
        Integer.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          emptyIndexValue, searchable, readPermission, writePermission, width,
          height, maxChars, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations, readonly,
          alignOnEdit, autoComplete)
        self.pythonType = float

    def getFormattedValue(self, obj, value, layout='view', showChanges=False,
                          language=None):
        return utils.formatNumber(value, sep=self.sep[0],
                                  precision=self.precision, tsep=self.tsep)

    def replaceSeparators(self, value):
        '''Replaces, in p_value, separators "sep" and "tsep" in such a way that
           p_value may become a valid Python float literal.'''
        for sep in self.sep: value = value.replace(sep, '.')
        return value.replace(self.tsep, '')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import copy
from appy.utils import sequenceTypes
from appy.model.workflow import Role

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
PERMISSION_NOT_FOUND = 'permission "%s" is not among permissions dict for ' \
  'state %s in workflow %s.'
                            
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class State:
    '''Represents a workflow state'''
    def __init__(self, permissions, initial=False, phase=None):
        self.usedRoles = {}
        # Dict "permissions" lists, for every permission managed by a workflow,
        # the roles for which the permission is granted in this state. Standard
        # permissions are 'read', 'write' and 'delete'. p_permissions must be a
        # dict with format
        #            ~{s_permissionName:[s_roleName|Role_role]}~

        # It will be converted by m_standardizeRoles into format
        #            ~{s_permissionName:{s_roleName:Role_role}}~
        self.permissions = permissions
        self.initial = initial
        self.phase = phase
        # Standardize the way roles are expressed within self.permissions
        self.standardizeRoles()

    def init(self, workflow, name):
        '''Lazy initialisation'''
        self.workflow = workflow
        self.name = name
        self.labelId = '%s_%s' % (workflow.name, name)

    def __repr__(self):
        return '<state %s::%s>' % (self.workflow.name, self.name)

    def copyPerms(self):
        '''Gets a deep copy of this state's permissions dict'''
        return copy.deepcopy(self.permissions)

    def getRole(self, role):
        '''p_role can be the name of a role or a Role instance. If it is the
           name of a role, this method returns self.usedRoles[role] if it
           exists, or creates a Role instance, puts it in self.usedRoles and
           returns it else. If it is a Role instance, the method stores it in
           self.usedRoles if it is not in it yet and returns it.'''
        if isinstance(role, str):
            if role in self.usedRoles:
                return self.usedRoles[role]
            else:
                theRole = Role(role)
                self.usedRoles[role] = theRole
                return theRole
        else:
            if role.name not in self.usedRoles:
                self.usedRoles[role.name] = role
            return role

    def standardizeRoles(self):
        '''This method converts, within self.permissions, every role to a
           Role instance. Every used role is stored in self.usedRoles.'''
        for permission, roles in self.permissions.items():
            if isinstance(roles, str) or isinstance(roles, Role):
                role = self.getRole(roles)
                self.permissions[permission] = {role.name: role}
            elif isinstance(roles, dict):
                for name, role in roles.iteritems():
                    roles[name] = self.getRole(role)
            else:
                # "roles" is a list or tuple, or None (nobody may have this
                # permission).
                d = {}
                if roles is not None:
                    for role in roles:
                        role = self.getRole(role)
                        d[role.name] = role
                self.permissions[permission] = d

    def getUsedRoles(self): return self.usedRoles.values()

    def getRolesFor(self, permission):
        '''Gets the roles that are granted p_permission on this state. r_ is a
           dict ~{s_roleName: Role}.'''
        if permission not in self.permissions:
            raise Exception(PERMISSION_NOT_FOUND % (permission, self.name, \
                                                    self.workflow.name))
        return self.permissions[permission]

    def addRoles(self, roles, permissions=()):
        '''Adds p_roles in self.permissions. p_roles can be a role name, a Role
           instance or a list of names and/or Role instances. If p_permissions
           is specified, roles are added to those permissions only. Else, roles
           are added for every permission within self.permissions.'''
        # Standardize parameters
        if type(roles) not in sequenceTypes: roles = (roles,)
        if isinstance(permissions, str): permissions = (permissions,)
        for perm, existingRoles in self.permissions.items():
            if permissions and (perm not in permissions): continue
            for role in roles:
                # Do nothing if "role" is already among existing roles
                name = role if isinstance(role, str) else role.name
                if name in existingRoles: continue
                # Add the role for this permission
                existingRoles[name] = self.getRole(role)

    def removeRoles(self, roleNames, permissions=()):
        '''Removes p_roleNames within dict self.permissions. If p_permissions is
           specified, removal is restricted to those permissions. Else, removal
           occurs throughout the whole dict self.permissions.'''
        if isinstance(roleNames, str): roleNames = (roleNames,)
        if isinstance(permissions, str): permissions = (permissions,)
        for perm, roles in self.permissions.items():
            if permissions and (perm not in permissions): continue
            for name in roleNames:
                # Remove this role if present in roles for this permission
                if name in roles:
                    del roles[name]

    def setRoles(self, roleNames, permissions=()):
        '''Sets p_rolesNames for p_permissions if not empty, for every
           permission in self.permissions else.'''
        if isinstance(roleNames, str): roleNames = (roleNames,)
        if isinstance(permissions, str): permissions = (permissions,)
        for perm in self.permissions:
            if permissions and (perm not in permissions): continue
            roles = self.permissions[perm] = {}
            for name in roleNames:
                roles[name] = self.getRole(name)

    def replaceRole(self, oldRoleName, newRoleName, permissions=()):
        '''Replaces p_oldRoleName by p_newRoleName. If p_permissions is
           specified, the replacement is restricted to those permissions. Else,
           replacements apply to the whole dict self.permissions.'''
        if isinstance(permissions, str): permissions = (permissions,)
        for perm, roles in self.permissions.items():
            if permissions and (perm not in permissions): continue
            # Find and replace p_oldRoleName with p_newRoleName
            if oldRoleName in roles:
                del roles[oldRoleName]
                roles[newRoleName] = self.getRole(newRoleName)

    def copyRoles(self, sourcePermission, destPermission):
        '''Overrides p_destPermission's roles with (a deep copy of)
           p_sourcePermission's roles.'''
        copiedRoles = copy.deepcopy(self.permissions[sourcePermission])
        self.permissions[destPermission] = copiedRoles

    def isIsolated(self):
        '''Returns True if, from this state, we cannot reach another state.
           Modifying a workflow for getting a state with auto-transitions only
           is a common technique for disabling a state in a workflow.'''
        if self.initial: return
        for tr in self.workflow.transitions.values():
            # Ignore transitions that do not touch this state
            if not tr.hasState(self, True) and not tr.hasState(self, False):
                continue
            # Transition "tr" has this state as start or end state. If start and
            # end states are different, it means that the state is not
            # isolated.
            if tr.isSingle():
                for state in tr.states:
                    if state != self: return
            else:
                for start, end in tr.states:
                    # Bypass (start, end) pairs having nothing to do with self
                    if (start != self) and (end != self): continue
                    if (start != self) or (end != self): return
        # If we are here, either there was no transition starting from self,
        # either all transitions were auto-transitions: self is then isolated.
        return True
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Injection of external content into XHTML chunks from Rich fields'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import re

from appy.xml.escape import Escape
from appy.utils.pretty import PrettyPrinter

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
INJECT_T_KO = 'Wrong injection type "p%s".'
PX_CTX_KO   = 'Cannot render "%s" - PX context is absent or incomplete.'
PX_KO       = 'Error while executing PX "%s": %s'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Injector:
    '''Implements the "injection" mechanism in a Rich field'''

    # More information on field "inject", defined in class
    # appy.model.fields.rich.Rich.

    # Injector-specific error class
    class Error(Exception): pass

    # Regular expression allowing to identify, within a Rich value, an injection
    # link.
    injectLink = re.compile('<a href="(.*?)">p([xy]):\s*(.*?)\s*</a>')

    @classmethod
    def executePx(class_, specifier, o):
        '''Executes the PX denoted by p_specifier'''
        # It is not possible to find the current PX context, if there is no
        # p_o(bject).
        if o is None: return ''
        # It is not possible to execute the PX if the current PX context is
        # absent or incomplete.
        try:
            Px = o.traversal.context.Px
            if Px is None:
                raise class_.Error(PX_CTX_KO % specifier)
        except AttributeError as err:
            raise class_.Error(PX_CTX_KO % specifier)
        # Execute the PX
        try:
            return Px.callFromSpec(specifier, o)
        except Px.Error as err:
            raise class_.Error(PX_KO % (specifier, str(err)))

    @classmethod
    def pxError(class_, message):
        '''Returns this PX-related error m_message, wrapped in a "pre" tag'''
        return '<pre>%s</pre>' % Escape.xhtml(message)

    @classmethod
    def getContent(class_, url, type, specifier, o=None):
        '''A link has been found in a chunk of XHTML code, with this p_url, of
           this p_type ("x" meaning "px:", "y" meaning "py:"), and this
           p_specifier (=the part following "px:" or "py:" in the tag's
           content). This method returns the corresponding content.'''
        if type == 'y':
            # A link to an external Python source code file
            r = PrettyPrinter.getFromUrl(url, specifier, o=o)
        elif type == 'x':
            # p_specifier denotes a PX. Get it and call it.
            try:
                r = class_.executePx(specifier, o)
            except class_.Error as err:
                r = class_.pxError(str(err))
        else:
            r = class_.pxError(INJECT_T_KO % type)
        return r

    @classmethod
    def run(class_, xhtml, o=None):
        '''Perform Rich field injections within this chunk of p_xhtml code'''
        return class_.injectLink.sub(
                 lambda m: class_.getContent(*m.groups(), o=o), xhtml)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy import Object, commercial
from appy.pod import PodError
from appy.shared.utils import Traceback
from appy.shared.errors import CommercialError
from appy.pod.elements import *

# ------------------------------------------------------------------------------
EVAL_ERROR = 'Error while evaluating expression "%s". %s'
FROM_EVAL_ERROR = 'Error while evaluating the expression "%s" defined in the ' \
                  '"from" part of a statement. %s'
WRONG_SEQ_TYPE = 'Expression "%s" is not iterable.'
WRONG_ITERATOR = 'Name "%s" cannot be used for an iterator variable.'
TABLE_NOT_ONE_CELL = "The table you wanted to populate with '%s' " \
                     "can\'t be dumped with the '-' option because it has " \
                     "more than one cell in it."

class EvaluationError(Exception):
    def __init__(self, originalError, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)
        self.originalError = originalError

# ------------------------------------------------------------------------------
class Action:
    '''Abstract class representing a action (=statement) that must be performed
       on the content of a buffer (if, for...).'''

    def __init__(self, name, buffer, expr, elem, minus):
        # Actions may be named. Currently, the name of an action is only used
        # for giving a name to "if" actions; thanks to this name, "else" actions
        # that are far away may reference their "if".
        self.name = name
        # The buffer hosting the action
        self.buffer = buffer
        # The Python expression to evaluate (may be None in the case of a
        # Null or Else action, for example).
        self.expr = expr
        # The element within the buffer that is the action's target
        self.elem = elem
        # If "minus" is True, the main elem(s) must not be dumped
        self.minus = minus
        # If "source" is "buffer", we must dump the (evaluated) buffer content.
        # If it is 'from', we must dump what comes from the "from" part of the
        # action (='fromExpr'). See m_setFrom below.
        self.source = 'buffer'
        self.fromExpr = self.fromPlus = None
        # Several actions may co-exist for the same buffer, as a chain of Action
        # instances, defined via the following attribute.
        self.subAction = None

    def setFrom(self, plus, expr):
        '''Associate to this action a "from" clause (pod only)'''
        self.source = 'from'
        self.fromPlus = plus
        self.fromExpr = expr

    def getExceptionLine(self, e):
        '''Gets the line describing exception p_e, containing the exception
           class, message and line number.'''
        return '%s: %s' % (e.__class__.__name__, str(e))

    def manageError(self, result, context, errorMessage, originalError=None):
        '''Manage the encountered error: dump it into the buffer or raise an
           exception.'''
        if self.buffer.env.raiseOnError:
            if not self.buffer.pod:
                # Add in the error message the line nb where the errors occurs
                # within the PX.
                locator = self.buffer.env.parser.locator
                # The column number may not be given
                col = locator.getColumnNumber()
                if col is None: col = ''
                else: col = ', column %d' % col
                errorMessage += ' (line %s%s)' % (locator.getLineNumber(), col)
                # Integrate the traceback (at least, its last lines)
                errorMessage += '\n' + Traceback.get(6).decode('utf-8')
            if originalError:
                raise EvaluationError(originalError, errorMessage)
            raise Exception(errorMessage)
        # Create a temporary buffer to dump the error. If I reuse this buffer to
        # dump the error (what I did before), and we are, at some depth, in a
        # for loop, this buffer will contain the error message and not the
        # content to repeat anymore. It means that this error will also show up
        # for every subsequent iteration.
        tempBuffer = self.buffer.clone()
        PodError.dump(tempBuffer, errorMessage, withinElement=self.elem)
        tempBuffer.evaluate(result, context)

    def _evalExpr(self, expr, context):
        '''Evaluates p_expr with p_context. p_expr can contain an error expr,
           in the form "someExpr|errorExpr". If it is the case, if the "normal"
           expr raises an error, the "error" expr is evaluated instead.'''
        if '|' not in expr:
            res = eval(expr, context)
        else:
            expr, errorExpr = expr.rsplit('|', 1)
            try:
                res = eval(expr, context)
            except Exception:
                res = eval(errorExpr, context)
        return res

    def evaluateExpression(self, result, context, expr):
        '''Evaluates expression p_expr with the current p_context. Returns a
           tuple (result, errorOccurred).'''
        try:
            res = self._evalExpr(expr, context)
            error = False
        except Exception, e:
            # Hack for MessageException instances: always re-raise it as is
            if e.__class__.__name__ == 'MessageException': raise e
            res = None
            line = self.getExceptionLine(e)
            try:
                errorMessage = EVAL_ERROR % (expr, line)
            except UnicodeDecodeError:
                errorMessage = EVAL_ERROR % (expr, line.decode('utf-8'))
            self.manageError(result, context, errorMessage, e)
            error = True
        return res, error

    def execute(self, result, context):
        '''Executes this action given some p_context and add the result to
           p_result.'''
        # Check that if minus is set, we have an element which can accept it
        if self.minus and isinstance(self.elem, Table) and \
           (not self.elem.tableInfo.isOneCell()):
            self.manageError(result, context, TABLE_NOT_ONE_CELL % self.expr)
        else:
            error = False
            # Evaluate self.expr in eRes
            eRes = None
            if self.expr:
                eRes,error = self.evaluateExpression(result, context, self.expr)
            if not error:
                # Trigger action-specific behaviour
                self.do(result, context, eRes)

    def evaluateBuffer(self, result, context,
                       forceSource=None, ignoreMinus=False):
        '''Evaluates the buffer tied to this action and add the result in
           p_result. The source for evaluation can be forced to p_forceSource
           but in most cases depends on self.source.'''
        # Determine the source
        source = forceSource or self.source
        # Determine "minus"
        if ignoreMinus:
            minus = False
        else:
            minus = self.minus
        if source == 'buffer':
            self.buffer.evaluate(result, context, removeMainElems=minus)
        else:
            # Evaluate self.fromExpr in fromRes
            fromRes = None
            error = False
            try:
                fromRes = eval(self.fromExpr, context)
            except Exception, e:
                msg = FROM_EVAL_ERROR % (self.fromExpr,self.getExceptionLine(e))
                self.manageError(result, context, msg, e)
                error = True
            if not error:
                if not self.fromPlus:
                    # Write the result
                    result.write(fromRes)
                else:
                    # We must keep the root tag within self.buffer and dump the
                    # result into it.
                    content = self.buffer.content
                    result.write(content[:content.find('>') + 1])
                    result.write(fromRes)
                    result.write(content[content.rfind('<'):])

    def addSubAction(self, action):
        '''Adds p_action as a sub-action of this action'''
        if not self.subAction:
            self.subAction = action
            # Transmit "minus" to the sub-action. Indeed, the responsiblity to
            # dump content in the buffer is delegated to the sub-action,
            # "minus-ity" included.
            action.minus = self.minus
        else:
            self.subAction.addSubAction(action)

    def check(self):
        '''Returns a tuple (success, message) indicating if the action is well
           formed or not.'''
        return True, None

class If(Action):
    '''Action that determines if we must include the content of the buffer in
       the result or not.'''
    def do(self, result, context, exprRes):
        if exprRes:
            if self.subAction:
                self.subAction.execute(result, context)
            else:
                self.evaluateBuffer(result, context)
        else:
            if self.buffer.isMainElement(Cell.OD):
                # Don't leave the current row with a wrong number of cells
                result.dumpElement(Cell.OD.elem)

class Else(If):
    '''Action that is linked to a previous "if" action. In fact, an "else"
       action works exactly like an "if" action, excepted that instead of
       defining a conditional expression, it is based on the negation of the
       conditional expression of the last defined "if" action.'''

    def __init__(self, name, buff, expr, elem, minus, ifAction):
        If.__init__(self, name, buff, None, elem, minus)
        self.ifAction = ifAction

    def do(self, result, context, exprRes):
        # This action is executed if the tied "if" action is not executed
        ifAction = self.ifAction
        iRes, error = ifAction.evaluateExpression(result,context,ifAction.expr)
        If.do(self, result, context, not iRes)

class For(Action):
    '''Actions that will include the content of the buffer as many times as
       specified by the action parameters.'''

    def __init__(self, name, buff, expr, elem, minus, iters):
        Action.__init__(self, name, buff, expr, elem, minus)
        # Name of the iterator variable(s) used in each loop
        self.iters = iters

    def initialiseLoop(self, context, elems):
        '''Initialises information about the loop, before entering into it. It
           is possible that this loop overrides an outer loop whose iterator
           has the same name. This method returns a tuple
           (loop, outerOverriddenLoop).'''
        # The "loop" object, made available in the POD context, contains info
        # about all currently walked loops. For every walked loop, a specific
        # object, accessible at getattr(loop, self.iters[0]), stores info about
        # its status:
        # ----------------------------------------------------------------------
        #  length   | the total number of walked elements within the loop
        # ----------------------------------------------------------------------
        #  nb       | the index (starting at 0) of the currently walked element
        # ----------------------------------------------------------------------
        #  first    | True if the currently walked element is the first one
        # ----------------------------------------------------------------------
        #  last     | True if the currently walked element is the last one
        # ----------------------------------------------------------------------
        #  odd      | True if the currently walked element is odd
        # ----------------------------------------------------------------------
        #  even     | True if the currently walked element is even
        # ----------------------------------------------------------------------
        #  previous | Points to the previous element, if any
        # ----------------------------------------------------------------------
        # For example, if you have a "for" statement like this:
        #                 for elem in myListOfElements
        # Within the part of the ODT document impacted by this statement, you
        # may access to loop.elem.length to know the total length of
        # myListOfElements, or loop.elem.nb to know the index of the current
        # elem within myListOfElements.
        if 'loop' not in context:
            # Attribute "_all_" stores the list of all currently running loops
            loops = context['loop'] = Object(_all_=[])
        else:
            loops = context['loop']
        try:
            total = len(elems)
        except Exception:
            total = 0
        curLoop = Object(length=total, previous=None, buffer=self.buffer)
        loops._all_.append(curLoop)
        # Does this loop override an outer loop with homonym iterator ?
        outerLoop = None
        iter = self.iters[0]
        if hasattr(loops, iter):
            outerLoop = getattr(loops, iter)
        # Put this loop in the global object "loop"
        setattr(loops, iter, curLoop)
        return curLoop, outerLoop

    def updateContext(self, context, item, forcedValue=None):
        '''We are in the loop, and p_item is the currently walked item. We must
           update the context by adding or updating values for iterator
           variable(s).'''
        # In most cases, there is a single iterator variable: for x in list
        names = self.iters
        if len(names) == 1:
            if forcedValue is None:
                value = item
            else:
                value = forcedValue
            context[names[0]] = value
        # This is the case: for a, b, c in list
        else:
            i = 0
            while i < len(names):
                if forcedValue is None:
                    value = item[i]
                else:
                    value = forcedValue
                context[names[i]] = value
                i += 1

    def do(self, result, context, elems):
        '''Performs the "for" action. p_elems is the list of elements to
           walk, evaluated from self.expr.'''
        # Check p_exprRes type
        try:
            # All "iterable" objects are OK
            iter(elems)
        except TypeError, te:
            self.manageError(result, context, WRONG_SEQ_TYPE % self.expr, te)
            return
        # Remember variables hidden by iterators if any
        hiddenVars = {}
        for name in self.iters:
            # Prevent reserved names to be used
            if name == '_all_':
                self.manageError(result, context, WRONG_ITERATOR % name)
                return
            if name in context:
                hiddenVars[name] = context[name]
        # In the case of cells, initialize some values
        isCell = False
        if isinstance(self.elem, Cell):
            isCell = True
            if 'columnsRepeated' in context:
                # This feature is only available in the open source version
                if commercial: raise CommercialError()
                nbOfColumns = sum(context['columnsRepeated'])
                customColumnsRepeated = True
            else:
                nbOfColumns = self.elem.tableInfo.nbOfColumns
                customColumnsRepeated = False
            initialColIndex = self.elem.colIndex
            currentColIndex = initialColIndex
            rowAttributes = self.elem.tableInfo.curRowAttrs
            # If p_elems is empty, dump an empty cell to avoid having the wrong
            # number of cells for the current row.
            if not elems:
                result.dumpElement(Cell.OD.elem)
        # Enter the "for" loop
        loop, outerLoop = self.initialiseLoop(context, elems)
        i = -1
        for item in elems:
            i += 1
            loop.nb = i
            loop.first = i == 0
            loop.last = i == (loop.length-1)
            loop.even = (i%2)==0
            loop.odd = not loop.even
            self.updateContext(context, item)
            # Cell: add a new row if we are at the end of a row
            if isCell and (currentColIndex == nbOfColumns):
                result.dumpEndElement(Row.OD.elem)
                result.dumpStartElement(Row.OD.elem, rowAttributes)
                currentColIndex = 0
            # If a sub-action is defined, execute it
            if self.subAction:
                self.subAction.execute(result, context)
            else:
                # Evaluate the buffer directly
                self.evaluateBuffer(result, context)
            # Cell: increment the current column index
            if isCell:
                currentColIndex += 1
            loop.previous = item
        # Cell: leave the last row with the correct number of cells, excepted
        # if the user has specified himself "columnsRepeated": it is his
        # responsibility to produce the correct number of cells.
        if isCell and elems and not customColumnsRepeated:
            wrongNbOfCells = (currentColIndex-1) - initialColIndex
            if wrongNbOfCells < 0: # Too few cells for last row
                for i in range(abs(wrongNbOfCells)):
                    self.updateContext(context, None, forcedValue='')
                    self.buffer.evaluate(result, context, subElements=False)
                    # This way, the cell is dumped with the correct styles
            elif wrongNbOfCells > 0: # Too many cells for last row
                # Finish current row
                nbOfMissingCells = 0
                if currentColIndex < nbOfColumns:
                    nbOfMissingCells = nbOfColumns - currentColIndex
                    self.updateContext(context, None, forcedValue='')
                    for i in range(nbOfMissingCells):
                        self.buffer.evaluate(result, context, subElements=False)
                result.dumpEndElement(Row.OD.elem)
                # Create additional row with remaining cells
                result.dumpStartElement(Row.OD.elem, rowAttributes)
                nbOfRemainingCells = wrongNbOfCells + nbOfMissingCells
                nbOfMissingCellsLastLine = nbOfColumns - nbOfRemainingCells
                self.updateContext(context, None, forcedValue='')
                for i in range(nbOfMissingCellsLastLine):
                    self.buffer.evaluate(result, context, subElements=False)
        # Delete the current loop object and restore the overridden one if any
        loops = context['loop']
        name = self.iters[0]
        try:
            delattr(loops, name)
        except AttributeError:
            pass
        if outerLoop:
            setattr(loops, name, outerLoop)
        loops._all_.pop()
        # Restore hidden variables and remove iterator variables from the
        # context.
        context.update(hiddenVars)
        if elems:
            for name in self.iters:
                if (name not in hiddenVars) and (name in context):
                    # On error, name may not be in the context
                    del context[name]

class Null(Action):
    '''Action that does nothing. Used in conjunction with a "from" clause, it
       allows to insert in a buffer arbitrary odt content.'''
    noFromError = 'There was a problem with this action. Possible causes: ' \
      '(1) you specified no action (ie "do text") while not specifying any ' \
      'from clause; (2) you specified the from clause on the same line as ' \
      'the action, which is not allowed (ie "do text from ...").'

    def __init__(self, buff, elem):
        Action.__init__(self, '', buff, None, elem, None)

    def do(self, result, context, exprRes):
        self.evaluateBuffer(result, context)

    def check(self):
        '''This action must have a tied from clause'''
        if self.source != 'from':
            return False, self.noFromError
        return True, None

class MetaIf(Action):
    '''Action allowing a note not to be evaluated and re-dumped as-is in the
       result, depending on some (meta-) condition.'''

    def __init__(self, buff, subExpr, elem, minus, statements):
        Action.__init__(self, '', buff, subExpr, elem, minus)
        # The list of statements containing in the original note
        self.statements = statements

    def reifyNote(self):
        '''Recreate the note as close to the original as possible'''
        # Use some fake buffer and dump the note in it. Reuse the code for
        # dumping an error, also dumped as a note.
        buf = self.buffer.clone()
        PodError.dump(buf, u'</text:p>\n<text:p>'.join(self.statements),
                      dumpTb=False, escapeMessage=False)
        return buf

    def do(self, result, context, exprRes):
        if exprRes:
            # The "meta-if" condition is True. It means that content must really
            # be dumped.
            if self.subAction:
                self.subAction.execute(result, context)
            else:
                self.evaluateBuffer(result, context)
        else:
            # The note must be dumped unevaluated in the result; the potential
            # "from" expression must not be evaluated as well.
            note = self.reifyNote()
            # Inject the note in the buffer, after the first 'text:p' tag: else,
            # it might not be rendered. self.minus must not be interpreted. When
            # reifying several notes in the same paragraph, ensure they are
            # reified in order, by forcing to insert a note after another one.
            self.buffer.insertSubBuffer(note, after='text:p',
                                        afterClosing='office:annotation')
            self.evaluateBuffer(result, context,
                                forceSource='buffer', ignoreMinus=True)

class Variables(Action):
    '''Action that allows to define a set of variables somewhere in the
       template.'''
    def __init__(self, name, buff, elem, minus, variables):
        # We do not use the default Buffer.expr attribute for storing the Python
        # expression, because here we will have several expressions, one for
        # every defined variable.
        Action.__init__(self, name, buff, None, elem, minus)
        # Definitions of variables: ~[(s_name|[s_name], s_expr)]~
        self.variables = variables

    def storeVariable(self, name, value, context, hidden):
        '''Adds a variable named p_name with this p_value in the p_context.
           Updates the dict of p_hidden variables and r_eturn it.'''
        if name.startswith('@'):
            # "name" represents a global variable. Update its value in the
            # context with p_value.
            context[name[1:]] = value
        else:
            # Store variable p_name with p_value in the p_context. If this
            # variable already exists in the context, remember its previous
            # value in p_hidden.
            if name in context:
                if not hidden:
                    hidden = {name: context[name]}
                else:
                    hidden[name] = context[name]
            # Store the result into the context
            context[name] = value
        return hidden

    def removeVariable(self, name, context, hidden):
        '''Remove, when relevant, variable p_name from the p_context'''
        # Do not remove it if it is a global variable
        if name.startswith('@'): return
        # Do not remove it if it is an overridden variable
        if hidden and (name in hidden): return
        del context[name]

    def do(self, result, context, exprRes):
        '''Evaluate the variables' expressions: because there are several
           expressions, we do not use the standard, single-expression-minded
           Action code for evaluating expressions.

           We remember the names and values of the variables that we will hide
           in the context: after execution of this buffer, we will restore those
           values.
        '''
        hidden = None
        for names, expr in self.variables:
            # Evaluate variable expression in v_value
            value, error = self.evaluateExpression(result, context, expr)
            if error: return
            if isinstance(names, basestring):
                # A single variable name
                hidden = self.storeVariable(names, value, context, hidden)
            else:
                # There are several variables whose values must be initialized
                # by unpacking v_values.
                i = -1
                for name in names:
                    i += 1
                    hidden = self.storeVariable(name, value[i], context, hidden)
        # If a sub-action is defined, execute it
        if self.subAction:
            self.subAction.execute(result, context)
        else:
            # Evaluate the buffer directly
            self.evaluateBuffer(result, context)
        # Restore hidden variables if any
        if hidden: context.update(hidden)
        # Delete not-hidden variables
        for names, expr in self.variables:
            if isinstance(names, basestring):
                self.removeVariable(names, context, hidden)
            else:
                for name in names:
                    self.removeVariable(name, context, hidden)
# ------------------------------------------------------------------------------
